/* c_x11.c
** This  file  is   free  software,  which   comes  along  with   thread-safe
** SmallEiffel. This software  is distributed  in the  hope that  it will  be
** useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
** you want, provided this  header is kept unaltered,  and a notification  of
** the changes is  added. You  are allowed to  redistribute it  and sell  it,
** alone or as a part of another product.
**
** (C) Miguel Oliveira e Silva, September 2003
*/

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "c_x11.h"

typedef struct
{
  XColor color;
  unsigned long pixel; /* pixel value */
} X11Color;

typedef struct
{
  Window window;
  Display *display;
  Window parent;
  XSizeHints hints;
  XWindowAttributes attributes;
} X11Window;

typedef struct
{
  Display *display;
  Font font;
  XFontStruct *fs;
} X11Font;

typedef struct
{
  char *display_name;
  Display *display;
  int screen;
  Window root_window;
  X11Color *black_color;
  X11Color *white_color;
  XEvent event;
  Colormap default_colormap;
  KeySym key;
  XComposeStatus cs;
  char key_buffer[20];
} X11Display;

static int initialized = 0;
static int threads_usable;

void x_initialize(void)
{
  threads_usable = (XInitThreads() ? 1 : 0);
  initialized = 1;
}

void x_terminate(void)
{
  initialized = 0;
}

int x_initialized(void)
{
  return initialized;
}

int x_threads_usable(void)
{
  return threads_usable;
}

X11Display *x_open_display_with_name(char *display_name)
  /* returns NULL on failure */
{
  X11Display *result;

  result = (X11Display *)malloc(sizeof(X11Display));
  assert(result != NULL);

  if (display_name != "")
  {
    result->display_name = (char *)malloc(strlen(display_name) + 1);
    strcpy(result->display_name,display_name);
  }
  else
    result->display_name = "";

  result->display = XOpenDisplay(display_name);
  if (result->display != NULL)
  {
    result->screen = XDefaultScreen(result->display);
    result->root_window = XDefaultRootWindow(result->display);
    result->black_color = (X11Color *)malloc(sizeof(X11Color));
    assert(result->black_color != NULL);
    result->black_color->pixel = XBlackPixel(result->display,result->screen);
    result->white_color = (X11Color *)malloc(sizeof(X11Color));
    assert(result->white_color != NULL);
    result->white_color->pixel = XWhitePixel(result->display,result->screen);
    result->default_colormap = DefaultColormap(result->display,result->screen);
  }
  else
  {
    if (display_name != "")
      free(result->display_name);
    free(result);
    result = NULL;
  }

  return result;
}

X11Display *x_open_display(void)
  /* returns NULL on failure */
{
  return x_open_display_with_name("");
}

X11Display *x_close_display(X11Display *x11_disp)
{
  XCloseDisplay(x11_disp->display);
  free(x11_disp);

  return NULL;
}

void x_lock_display(X11Display *x11_disp)
{
  XLockDisplay(x11_disp->display);
}

void x_unlock_display(X11Display *x11_disp)
{
  XUnlockDisplay(x11_disp->display);
}

void x_next_event(X11Display *x11_disp)
{
  XNextEvent(x11_disp->display,&x11_disp->event);
  if(x11_disp->event.type == MappingNotify)
    XRefreshKeyboardMapping(&x11_disp->event.xmapping);
  else if (x11_disp->event.type == KeyPress || x11_disp->event.type == KeyRelease)
  {
    int c;

    c = XLookupString(&x11_disp->event.xkey,x11_disp->key_buffer,20,&x11_disp->key,&x11_disp->cs);
    x11_disp->key_buffer[c] = '\0';
  }
}

int x_last_event(X11Display *x11_disp)
{
  return x11_disp->event.type;
}

int x_is_last_event_window(X11Display *x11_disp,X11Window *x11_win)
{
  return (x11_disp->event.xany.window == x11_win->window ? 1 : 0);
}

int x_last_event_pointer_x_pos(X11Display *x11_disp)
{
  int result = 0;

  if (x11_disp->event.type == KeyPress ||
      x11_disp->event.type == KeyRelease)
    result = x11_disp->event.xkey.x;
  else if (x11_disp->event.type == ButtonPress ||
           x11_disp->event.type == ButtonRelease)
    result = x11_disp->event.xbutton.x;
  else if (x11_disp->event.type == MotionNotify)
    result = x11_disp->event.xmotion.x;

  return result;
}

int x_last_event_pointer_y_pos(X11Display *x11_disp)
{
  int result = 0;

  if (x11_disp->event.type == KeyPress ||
      x11_disp->event.type == KeyRelease)
    result = x11_disp->event.xkey.y;
  else if (x11_disp->event.type == ButtonPress ||
           x11_disp->event.type == ButtonRelease)
    result = x11_disp->event.xbutton.y;
  else if (x11_disp->event.type == MotionNotify)
    result = x11_disp->event.xmotion.y;

  return result;
}

char *x_keyboard_event_to_string(X11Display *x11_disp)
{
  return x11_disp->key_buffer;
}

KeySym x_keyboard_event_keysym(X11Display *x11_disp)
{
  return x11_disp->key;
}

/* Color (simple [incomplete] wrap) */

X11Color *x_make_white_pixel(X11Display *x11_disp)
{
  return x11_disp->white_color;
}

X11Color *x_make_black_pixel(X11Display *x11_disp)
{
  return x11_disp->black_color;
}

X11Color *x_make_color_pixel(X11Display *x11_disp,char *name)
{
  X11Color *result;
  XColor color;

  result = (X11Color *)malloc(sizeof(X11Color));
  assert(result != NULL);

  if(XAllocNamedColor(x11_disp->display,x11_disp->default_colormap,
                      name,&(result->color),&color))
  {
    result->pixel = color.pixel;
  }
  else
  {
    free(result);
    result = NULL;
  }

  return result;
}

/* Windows */

void x_get_window_attributes(X11Window *x11_win)
{
  XGetWindowAttributes(x11_win->display,x11_win->window,&x11_win->attributes);
}

X11Window *x_create_simple_window_with_parent(X11Display *x11_disp,
             X11Window *parent,
             int x,int y,unsigned int width,unsigned int height,
             unsigned int border_width,X11Color *border_color,
             X11Color *background_color)
{
  X11Window *result;

  result = (X11Window *)malloc(sizeof(X11Window));
  assert(result != NULL);

  result->display = x11_disp->display;

  if (parent == NULL)
    result->parent = x11_disp->root_window;
  else
    result->parent = parent->window;

  result->hints.x = x;
  result->hints.y = y;
  result->hints.width = width;
  result->hints.height = height;
  result->hints.flags = PPosition | PSize;

  result->window = XCreateSimpleWindow(x11_disp->display,result->parent,
		                       x,y,width,height,
                                       border_width,border_color->pixel,
                                       background_color->pixel);
  XSetStandardProperties(result->display,result->window,
                         "","",None,NULL,0,&result->hints);
  x_get_window_attributes(result);
  return result;
}

X11Window *x_create_simple_window(X11Display *x11_disp,
             int x,int y,unsigned int width,unsigned int height,
             unsigned int border_width,X11Color *border_color,
             X11Color *background_color)
{
  return x_create_simple_window_with_parent(x11_disp,NULL,x,y,width,height,
                        border_width,border_color,background_color);
}

X11Window *x_destroy_window(X11Display *x11_disp,X11Window *x11_win)
{
  XDestroyWindow(x11_disp->display,x11_win->window);

  return NULL;
}

void x_destroy_subwindows(X11Display *x11_disp,X11Window *x11_win)
{
  XDestroySubwindows(x11_disp->display,x11_win->window);
}

int x_get_window_origin_x(X11Window *x11_win)
{
  x_get_window_attributes(x11_win);
  return x11_win->attributes.x;
}

int x_get_window_origin_y(X11Window *x11_win)
{
  x_get_window_attributes(x11_win);
  return x11_win->attributes.y;
}

int x_get_window_origin_width(X11Window *x11_win)
{
  x_get_window_attributes(x11_win);
  return x11_win->attributes.width;
}

int x_get_window_origin_height(X11Window *x11_win)
{
  x_get_window_attributes(x11_win);
  return x11_win->attributes.height;
}

int x_get_window_origin_border_width(X11Window *x11_win)
{
  x_get_window_attributes(x11_win);
  return x11_win->attributes.border_width;
}

int x_get_window_origin_depth(X11Window *x11_win)
{
  x_get_window_attributes(x11_win);
  return x11_win->attributes.depth;
}

void x_flush(X11Display *x11_disp)
{
  XFlush(x11_disp->display);
}

int x_display_height(X11Display *x11_disp)
{
  return DisplayHeight(x11_disp->display,x11_disp->screen);
}

int x_display_width(X11Display *x11_disp)
{
  return DisplayWidth(x11_disp->display,x11_disp->screen);
}

int x_display_height_mm(X11Display *x11_disp)
{
  return DisplayHeightMM(x11_disp->display,x11_disp->screen);
}

int x_display_width_mm(X11Display *x11_disp)
{
  return DisplayWidthMM(x11_disp->display,x11_disp->screen);
}

void x_map_raised(X11Display *x11_disp,X11Window *x11_win)
{
  XMapRaised(x11_disp->display,x11_win->window);
}

void x_map_window(X11Display *x11_disp,X11Window *x11_win)
{
  XMapWindow(x11_disp->display,x11_win->window);
}

void x_map_subwindows(X11Display *x11_disp,X11Window *x11_win)
{
  XMapSubwindows(x11_disp->display,x11_win->window);
}

void x_unmap_window(X11Display *x11_disp,X11Window *x11_win)
{
  XUnmapWindow(x11_disp->display,x11_win->window);
}

void x_unmap_subwindows(X11Display *x11_disp,X11Window *x11_win)
{
  XUnmapSubwindows(x11_disp->display,x11_win->window);
}

void x_raise_window(X11Display *x11_disp,X11Window *x11_win)
{
  XRaiseWindow(x11_disp->display,x11_win->window);
}

void x_lower_window(X11Display *x11_disp,X11Window *x11_win)
{
  XLowerWindow(x11_disp->display,x11_win->window);
}

void x_select_input(X11Window *x11_win,int mask)
{
  XSelectInput(x11_win->display,x11_win->window,(long)mask);
}


/* Graphic Contexts */

#define DrawableWindow 1

typedef struct
{
  GC gc;
  Display *display;
  Window window;
  X11Color *foreground_color;
  X11Color *background_color;
  XGCValues values;
} X11GC;

#define GCAllMask (GCFunction | GCPlaneMask | GCForeground | GCBackground | \
                   GCLineWidth | GCLineStyle | GCCapStyle | GCJoinStyle |  \
                   GCFillStyle | GCFillRule | GCTile | GCStipple |  \
                   GCTileStipXOrigin | GCTileStipYOrigin | GCFont | \
                   GCSubwindowMode | GCGraphicsExposures | GCClipXOrigin | \
                   GCClipYOrigin | GCClipMask | GCDashOffset | GCDashList | \
                   GCArcMode)

void x_set_foreground(X11GC *x11_gc,X11Color *color);
void x_set_background(X11GC *x11_gc,X11Color *color);

X11GC *x_create_gc(X11Display *x11_disp,X11Window *x11_win)
{
  X11GC *result = NULL;

  result = (X11GC *)malloc(sizeof(X11GC));
  assert(result != NULL);

  result->display = x11_disp->display;
  result->window = x11_win->window;
  result->gc = XCreateGC(result->display,result->window,0,0);
  XGetGCValues(result->display,result->gc,GCAllMask,&(result->values));
  x_set_foreground(result,x11_disp->black_color);
  x_set_background(result,x11_disp->white_color);

  return result;
}

X11GC *x_free_gc(X11GC *x11_gc)
{
  XFreeGC(x11_gc->display,x11_gc->gc);
  free(x11_gc);

  return NULL;
}

X11Color *x_foreground(X11GC *x11_gc)
{
  return x11_gc->foreground_color;
}

void x_set_foreground(X11GC *x11_gc,X11Color *color)
{
  x11_gc->foreground_color = color;
  XSetForeground(x11_gc->display,x11_gc->gc,color->pixel);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

X11Color *x_background(X11GC *x11_gc)
{
  return x11_gc->background_color;
}

void x_set_background(X11GC *x11_gc,X11Color *color)
{
  x11_gc->background_color = color;
  XSetBackground(x11_gc->display,x11_gc->gc,color->pixel);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

int x_function(X11GC *x11_gc)
{
  return x11_gc->values.function;
}

void x_set_function(X11GC *x11_gc,int function)
{
  XSetFunction(x11_gc->display,x11_gc->gc,function);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

int x_line_width(X11GC *x11_gc)
{
  return x11_gc->values.line_width;
}

int x_line_style(X11GC *x11_gc)
{
  return x11_gc->values.line_style;
}

int x_cap_style(X11GC *x11_gc)
{
  return x11_gc->values.cap_style;
}

int x_join_style(X11GC *x11_gc)
{
  return x11_gc->values.join_style;
}

void x_set_line_attributes(X11GC *x11_gc,int width,int style,int cap_style,int join_style)
{
  XSetLineAttributes(x11_gc->display,x11_gc->gc,width,style,cap_style,join_style);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

int x_arc_mode(X11GC *x11_gc)
{
  return x11_gc->values.arc_mode;
}

void x_set_arc_mode(X11GC *x11_gc,int arc_mode)
{
  XSetArcMode(x11_gc->display,x11_gc->gc,arc_mode);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

int x_fill_style(X11GC *x11_gc)
{
  return x11_gc->values.fill_style;
}

void x_set_fill_style(X11GC *x11_gc,int fill_style)
{
  XSetFillStyle(x11_gc->display,x11_gc->gc,fill_style);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

int x_fill_rule(X11GC *x11_gc)
{
  return x11_gc->values.fill_rule;
}

void x_set_fill_rule(X11GC *x11_gc,int fill_rule)
{
  XSetFillRule(x11_gc->display,x11_gc->gc,fill_rule);
  XGetGCValues(x11_gc->display,x11_gc->gc,GCAllMask,&(x11_gc->values));
}

void x_set_font(X11GC *x11_gc,X11Font *x11_font)
{
  XSetFont(x11_gc->display,x11_gc->gc,x11_font->font);
}

void x_clear_window(X11Display *x11_disp,X11Window *x11_win)
{
  XClearWindow(x11_disp->display,x11_win->window);
}

void x_clear_area(X11Display *x11_disp,X11Window *x11_win,int x,int y,int width,int height,int exposures)
{
  XClearArea(x11_disp->display,x11_win->window,x,y,width,height,exposures);
}


/* Drawing */

void x_draw_point(X11Window *x11_win,X11GC *x11_gc,int x,int y)
{
  XDrawPoint(x11_win->display,x11_win->window,x11_gc->gc,x,y);
}

#include <pthread.h>

#define INITIAL_XLIST_ARRAY_SIZE 32
#define INCREMENT_XLIST_ARRAY_SIZE 16

typedef struct
{
  XPoint *list;
  int num_points;
  int max_num_points;
} XPointStruct;

pthread_key_t point_list_tsk;
pthread_once_t point_list_tsk_once_ = PTHREAD_ONCE_INIT;

void point_list_tsk_once_routine(void)
{
  if (pthread_key_create(&point_list_tsk,NULL))
  {
    /* lacks appropriate error handling code !!! */
  }
}

XPointStruct *get_x_point_struct(void)
{
  XPointStruct *xps = NULL;

  pthread_once(&point_list_tsk_once_,point_list_tsk_once_routine);

  xps = (XPointStruct *)pthread_getspecific(point_list_tsk);

  if (xps == NULL)
  {
    xps = (XPointStruct *)malloc(sizeof(XPointStruct));
    assert(xps != NULL);
    if (pthread_setspecific(point_list_tsk,(void*)xps))
    {
      /* lacks appropriate error handling code !!! */
    }

    xps->max_num_points = INITIAL_XLIST_ARRAY_SIZE;
    xps->list = (XPoint *)malloc(xps->max_num_points * sizeof(XPoint));
    assert(xps->list != NULL);
    xps->num_points = 0;
  }
  else if (xps->num_points + 1 == xps->max_num_points)
  {
    xps->max_num_points += INCREMENT_XLIST_ARRAY_SIZE;
    xps->list = (XPoint *)realloc(xps->list,xps->max_num_points * sizeof(XPoint));
    assert(xps->list != NULL);
  }

  return xps;
}

void add_point_to_draw(int x,int y)
{
  XPointStruct *xps = NULL;

  xps = get_x_point_struct();
  xps->list[xps->num_points].x = x;
  xps->list[xps->num_points].y = y;
  xps->num_points++;
}

void clear_point_list(void)
{
  XPointStruct *xps = NULL;

  xps = get_x_point_struct();
  xps->num_points = 0;
}

void x_draw_points(X11Window *x11_win,X11GC *x11_gc,int mode)
{
  XPointStruct *xps = NULL;

  xps = get_x_point_struct();
  XDrawPoints(x11_win->display,x11_win->window,x11_gc->gc,xps->list,xps->num_points,mode);
}

void x_draw_lines(X11Window *x11_win,X11GC *x11_gc,int mode)
{
  XPointStruct *xps = NULL;

  xps = get_x_point_struct();
  XDrawLines(x11_win->display,x11_win->window,x11_gc->gc,xps->list,xps->num_points,mode);
}

void x_fill_polygon(X11Window *x11_win,X11GC *x11_gc,int shape,int mode)
{
  XPointStruct *xps = NULL;

  xps = get_x_point_struct();
  XFillPolygon(x11_win->display,x11_win->window,x11_gc->gc,xps->list,xps->num_points,shape,mode);
}


void x_draw_line(X11Window *x11_win,X11GC *x11_gc,int x1,int y1,int x2,int y2)
{
  XDrawLine(x11_win->display,x11_win->window,x11_gc->gc,x1,y1,x2,y2);
}


typedef struct
{
  XSegment *list;
  int num_segments;
  int max_num_segments;
} XSegmentStruct;

pthread_key_t segment_list_tsk;
pthread_once_t segment_list_tsk_once_ = PTHREAD_ONCE_INIT;

void segment_list_tsk_once_routine(void)
{
  if (pthread_key_create(&segment_list_tsk,NULL))
  {
    /* lacks appropriate error handling code !!! */
  }
}

XSegmentStruct *get_x_segment_struct(void)
{
  XSegmentStruct *xss = NULL;

  pthread_once(&segment_list_tsk_once_,segment_list_tsk_once_routine);

  xss = (XSegmentStruct *)pthread_getspecific(segment_list_tsk);

  if (xss == NULL)
  {
    xss = (XSegmentStruct *)malloc(sizeof(XSegmentStruct));
    assert(xss != NULL);
    if (pthread_setspecific(segment_list_tsk,(void*)xss))
    {
      /* lacks appropriate error handling code !!! */
    }

    xss->max_num_segments = INITIAL_XLIST_ARRAY_SIZE;
    xss->list = (XSegment *)malloc(xss->max_num_segments * sizeof(XSegment));
    assert(xss->list != NULL);
    xss->num_segments = 0;
  }
  else if (xss->num_segments + 1 == xss->max_num_segments)
  {
    xss->max_num_segments += INCREMENT_XLIST_ARRAY_SIZE;
    xss->list = (XSegment *)realloc(xss->list,xss->max_num_segments * sizeof(XSegment));
    assert(xss->list != NULL);
  }

  return xss;
}

void add_segment_to_draw(int x1,int y1,int x2,int y2)
{
  XSegmentStruct *xss = NULL;

  xss = get_x_segment_struct();
  xss->list[xss->num_segments].x1 = x1;
  xss->list[xss->num_segments].y1 = y1;
  xss->list[xss->num_segments].x2 = x2;
  xss->list[xss->num_segments].y2 = y2;
  xss->num_segments++;
}

void clear_segment_list(void)
{
  XSegmentStruct *xss = NULL;

  xss = get_x_segment_struct();
  xss->num_segments = 0;
}

void x_draw_segments(X11Window *x11_win,X11GC *x11_gc)
{
  XSegmentStruct *xss = NULL;

  xss = get_x_segment_struct();
  XDrawSegments(x11_win->display,x11_win->window,x11_gc->gc,xss->list,xss->num_segments);
}


void x_draw_arc(X11Window *x11_win,X11GC *x11_gc,int x,int y,int width,int height,int angle1, int angle2)
{
  XDrawArc(x11_win->display,x11_win->window,x11_gc->gc,x,y,width,height,angle1,angle2);
}


typedef struct
{
  XArc *list;
  int num_arcs;
  int max_num_arcs;
} XArcStruct;

pthread_key_t arc_list_tsk;
pthread_once_t arc_list_tsk_once_ = PTHREAD_ONCE_INIT;

void arc_list_tsk_once_routine(void)
{
  if (pthread_key_create(&arc_list_tsk,NULL))
  {
    /* lacks appropriate error handling code !!! */
  }
}

XArcStruct *get_x_arc_struct(void)
{
  XArcStruct *xas = NULL;

  pthread_once(&arc_list_tsk_once_,arc_list_tsk_once_routine);

  xas = (XArcStruct *)pthread_getspecific(arc_list_tsk);

  if (xas == NULL)
  {
    xas = (XArcStruct *)malloc(sizeof(XArcStruct));
    assert(xas != NULL);
    if (pthread_setspecific(arc_list_tsk,(void*)xas))
    {
      /* lacks appropriate error handling code !!! */
    }

    xas->max_num_arcs = INITIAL_XLIST_ARRAY_SIZE;
    xas->list = (XArc *)malloc(xas->max_num_arcs * sizeof(XArc));
    assert(xas->list != NULL);
    xas->num_arcs = 0;
  }
  else if (xas->num_arcs + 1 == xas->max_num_arcs)
  {
    xas->max_num_arcs += INCREMENT_XLIST_ARRAY_SIZE;
    xas->list = (XArc *)realloc(xas->list,xas->max_num_arcs * sizeof(XArc));
    assert(xas->list != NULL);
  }

  return xas;
}

void add_arc_to_draw(int x,int y,int width,int height,int angle1, int angle2)
{
  XArcStruct *xas = NULL;

  xas = get_x_arc_struct();
  xas->list[xas->num_arcs].x = x;
  xas->list[xas->num_arcs].y = y;
  xas->list[xas->num_arcs].width = width;
  xas->list[xas->num_arcs].height = height;
  xas->list[xas->num_arcs].angle1 = angle1;
  xas->list[xas->num_arcs].angle2 = angle2;
  xas->num_arcs++;
}

void clear_arc_list(void)
{
  XArcStruct *xas = NULL;

  xas = get_x_arc_struct();
  xas->num_arcs = 0;
}

void x_draw_arcs(X11Window *x11_win,X11GC *x11_gc)
{
  XArcStruct *xas = NULL;

  xas = get_x_arc_struct();
  XDrawArcs(x11_win->display,x11_win->window,x11_gc->gc,xas->list,xas->num_arcs);
}

void x_fill_arcs(X11Window *x11_win,X11GC *x11_gc)
{
  XArcStruct *xas = NULL;

  xas = get_x_arc_struct();
  XFillArcs(x11_win->display,x11_win->window,x11_gc->gc,xas->list,xas->num_arcs);
}


void x_fill_arc(X11Window *x11_win,X11GC *x11_gc,int x,int y,int width,int height,int angle1, int angle2)
{
  XFillArc(x11_win->display,x11_win->window,x11_gc->gc,x,y,width,height,angle1,angle2);
}

void x_draw_rectangle(X11Window *x11_win,X11GC *x11_gc,int x,int y,int width,int height)
{
  XDrawRectangle(x11_win->display,x11_win->window,x11_gc->gc,x,y,width,height);
}


typedef struct
{
  XRectangle *list;
  int num_rectangles;
  int max_num_rectangles;
} XRectangleStruct;

pthread_key_t rectangle_list_tsk;
pthread_once_t rectangle_list_tsk_once_ = PTHREAD_ONCE_INIT;

void rectangle_list_tsk_once_routine(void)
{
  if (pthread_key_create(&rectangle_list_tsk,NULL))
  {
    /* lacks appropriate error handling code !!! */
  }
}

XRectangleStruct *get_x_rectangle_struct(void)
{
  XRectangleStruct *xrs = NULL;

  pthread_once(&rectangle_list_tsk_once_,rectangle_list_tsk_once_routine);

  xrs = (XRectangleStruct *)pthread_getspecific(rectangle_list_tsk);

  if (xrs == NULL)
  {
    xrs = (XRectangleStruct *)malloc(sizeof(XRectangleStruct));
    assert(xrs != NULL);
    if (pthread_setspecific(rectangle_list_tsk,(void*)xrs))
    {
      /* lacks appropriate error handling code !!! */
    }

    xrs->max_num_rectangles = INITIAL_XLIST_ARRAY_SIZE;
    xrs->list = (XRectangle *)malloc(xrs->max_num_rectangles * sizeof(XRectangle));
    assert(xrs->list != NULL);
    xrs->num_rectangles = 0;
  }
  else if (xrs->num_rectangles + 1 == xrs->max_num_rectangles)
  {
    xrs->max_num_rectangles += INCREMENT_XLIST_ARRAY_SIZE;
    xrs->list = (XRectangle *)realloc(xrs->list,xrs->max_num_rectangles * sizeof(XRectangle));
    assert(xrs->list != NULL);
  }

  return xrs;
}

void add_rectangle_to_draw(int x,int y,int width,int height)
{
  XRectangleStruct *xrs = NULL;

  xrs = get_x_rectangle_struct();
  xrs->list[xrs->num_rectangles].x = x;
  xrs->list[xrs->num_rectangles].y = y;
  xrs->list[xrs->num_rectangles].width = width;
  xrs->list[xrs->num_rectangles].height = height;
  xrs->num_rectangles++;
}

void clear_rectangle_list(void)
{
  XRectangleStruct *xrs = NULL;

  xrs = get_x_rectangle_struct();
  xrs->num_rectangles = 0;
}

void x_draw_rectangles(X11Window *x11_win,X11GC *x11_gc)
{
  XRectangleStruct *xrs = NULL;

  xrs = get_x_rectangle_struct();
  XDrawRectangles(x11_win->display,x11_win->window,x11_gc->gc,xrs->list,xrs->num_rectangles);
}

void x_fill_rectangles(X11Window *x11_win,X11GC *x11_gc)
{
  XRectangleStruct *xrs = NULL;

  xrs = get_x_rectangle_struct();
  XFillRectangles(x11_win->display,x11_win->window,x11_gc->gc,xrs->list,xrs->num_rectangles);
}


void x_fill_rectangle(X11Window *x11_win,X11GC *x11_gc,int x,int y,int width,int height)
{
  XFillRectangle(x11_win->display,x11_win->window,x11_gc->gc,x,y,width,height);
}

void x_draw_string(X11Window *x11_win,X11GC *x11_gc,int x,int y,char *str,int length)
{
  XDrawString(x11_win->display,x11_win->window,x11_gc->gc,x,y,str,length);
}

void x_draw_image_string(X11Window *x11_win,X11GC *x11_gc,int x,int y,char *str,int length)
{
  XDrawImageString(x11_win->display,x11_win->window,x11_gc->gc,x,y,str,length);
}

X11Font *x_make_font(X11Display *x11_disp)
{
  X11Font *result;

  result = (X11Font *)malloc(sizeof(X11Font));
  assert(result != NULL);

  result->display = x11_disp->display;

  return result;
}

int x_font_exists(X11Font *x11_font,char *name)
{
  int result = 0,actual_count;
  char **list;

  list = XListFonts(x11_font->display,name,1,&actual_count);
  if (list != NULL)
  {
    result = 1;
    XFreeFontNames(list);
  }

  return result;
}

void x_load_font(X11Font *x11_font,char *name)
{
  x11_font->font = XLoadFont(x11_font->display,name);
  x11_font->fs = XQueryFont(x11_font->display,x11_font->font);
}

void x_unload_font(X11Font *x11_font)
{
  /* XUnloadFont(x11_font->display,x11_font->font); */
  XFreeFont(x11_font->display,x11_font->fs);
}

void x_text_extents(X11Font *x11_font,char *text,int text_length,
                    int *dir_hint,int *font_ascent,int *font_descent,
                    int *lbearing,int *rbearing,
                    int *width,int *ascent,int *descent)
{
  XCharStruct cf;

  XTextExtents(x11_font->fs,text,text_length,dir_hint,font_ascent,font_descent,
               &cf);
  *lbearing = cf.lbearing;
  *rbearing = cf.rbearing;
  *width = cf.width;
  *ascent = cf.ascent;
  *descent = cf.descent;
}

/* X11 cursors */

typedef struct
{
  Display *display;
  Cursor cursor;
} X11Cursor;

X11Cursor *x_make_cursor(X11Display *x11_disp)
{
  X11Cursor *result;

  result = (X11Cursor *)malloc(sizeof(X11Cursor));
  assert(result != NULL);

  result->display = x11_disp->display;

  return result;
}

void x_create_font_cursor(X11Cursor *x11_cur,unsigned int shape)
{
  x11_cur->cursor = XCreateFontCursor(x11_cur->display,shape);
}

void x_free_cursor(X11Cursor *x11_cur)
{
  XFreeCursor(x11_cur->display,x11_cur->cursor);
}

void x_define_cursor(X11Window *x11_win,X11Cursor *x11_cur)
{
  XDefineCursor(x11_win->display,x11_win->window,x11_cur->cursor);
}

void x_undefine_cursor(X11Window *x11_win)
{
  XUndefineCursor(x11_win->display,x11_win->window);
}

