-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, May 2003
--

class X11_DISPLAY

inherit
  X11_GLOBAL

creation
  initialize

feature

  initialize is
    require
      not initialized
    external "C"
    alias "x_initialize"
    end;

  terminate is
    require
      initialized
    external "C"
    alias "x_terminate"
    end;

  initialized: BOOLEAN is
    external "C"
    alias "x_initialized"
    end;

  threads_usable: BOOLEAN is
    require
      initialized
    external "C"
    alias "x_threads_usable"
    end;

feature

  open is
      -- Uses name of "DISPLAY" environment variable
    require
      not opened
    do
      disp_var := c_x_open_display;
      !!fetch_event_mutex.make;
      init_colors
    ensure
      not disp_var.is_null
    end;

  open_with_name(display_name: STRING) is
    require
      display_name /= Void;
      not opened
    do
      disp_var := c_x_open_display_with_name(display_name.to_external);
      !!fetch_event_mutex.make;
      init_colors
    ensure
      not disp_var.is_null
    end;

  opened: BOOLEAN is
    do
      Result := initialized and then not disp_var.is_null
    end;

  close is
    require
      opened
    do
      disp_var := c_x_close_display(disp_var)
    end;

  flush is
    require
      opened
    do
      c_x_flush(disp_var)
    end;

feature

  black_color: X11_COLOR;

  white_color: X11_COLOR;

  display_height_pixels: INTEGER is
    require
      opened
    do
      Result := c_x_display_height(disp_var)
    end;

  display_width_pixels: INTEGER is
    require
      opened
    do
      Result := c_x_display_width(disp_var)
    end;

  display_height_mm: INTEGER is
    require
      opened
    do
      Result := c_x_display_height_mm(disp_var)
    end;

  display_width_mm: INTEGER is
    require
      opened
    do
      Result := c_x_display_width_mm(disp_var)
    end;

feature -- locking

  lock is
    do
      c_x_lock_display(disp_var)
    end;

  unlock is
    do
      c_x_unlock_display(disp_var)
    end;

feature {X11_WINDOW} -- windows registering (for event lookup window)

  register_window(w: X11_WINDOW) is
    require
      w /= Void
    do
      list_windows.add_last(w)
    end;

  window_registed(w: X11_WINDOW): BOOLEAN is
    require
      w /= Void
    do
      Result := list_windows.fast_has(w)
    end;

feature -- events

  fetch_event is
    do
      fetch_event_mutex.lock;
      c_x_next_event(disp_var);
      onces.refresh("new_event");
      fetch_event_mutex.unlock
    end;

  last_event: INTEGER is
    do
      Result := c_x_last_event(disp_var)
    end;

  last_event_window: X11_WINDOW is
    local
      i: INTEGER
    once {"object","process","new_event"}
      fetch_event_mutex.lock;
      if internal_last_event_window = Void or else
         not c_x_is_last_event_window(disp_var,
               internal_last_event_window.win_var) then
        from
          i := list_windows.lower
        until
          i > list_windows.upper or else
          c_x_is_last_event_window(disp_var,(list_windows @ i).win_var)
        loop
          i := i + 1
        end;
        if not (i > list_windows.upper) then
          internal_last_event_window := list_windows @ i
        else
          internal_last_event_window := Void
        end
      end;
      Result := internal_last_event_window;
      fetch_event_mutex.unlock
    end;

  keyboard_event: BOOLEAN is
    do
      Result := last_event = event_key_press or
                last_event = event_key_release
    end;

  mouse_event: BOOLEAN is
    do
      Result := last_event = event_button_press or
                last_event = event_button_release or
                last_event = event_motion_notify
    end;

  pointer_x_position: INTEGER is
    require
      -- last event's pointer x position (relative to event window)
      mouse_event or keyboard_event
    do
      Result := c_x_last_event_pointer_x_pos(disp_var)
    end;

  pointer_y_position: INTEGER is
    require
      -- last event's pointer y position (relative to event window)
      mouse_event or keyboard_event
    do
      Result := c_x_last_event_pointer_y_pos(disp_var)
    end;

  keyboard_event_to_string: STRING is
    require
      keyboard_event
    do
      !!Result.from_external_copy(c_x_keyboard_event_to_string(disp_var))
    end;

  keyboard_event_to_string_in(str: STRING) is
      -- same as 'keyboard_event_to_string' but stores result in 'str'
    require
      keyboard_event
    do
      str.from_external_copy(c_x_keyboard_event_to_string(disp_var))
    end;

  key_sym: INTEGER is
    require
      keyboard_event
    do
      Result := c_x_keyboard_event_keysym(disp_var)
    end;

feature {X11_WINDOW,X11_COLOR,X11_GRAPHIC_CONTEXT,X11_FONT,X11_CURSOR}

  disp_var: POINTER;

feature {NONE}

  init_colors is
    do
      !!black_color.make_black_color(Current);
      !!white_color.make_white_color(Current)
    end;

  internal_last_event_window: X11_WINDOW;

  default_number_of_windows: INTEGER is 10; -- initial number of windows

  list_windows: ARRAY[X11_WINDOW] is
    once {"process","object"}
      !!Result.with_capacity(default_number_of_windows,0)
    end;

  fetch_event_mutex: MUTEX;

  c_x_open_display: POINTER is
    external "C"
    alias "x_open_display"
    end;

  c_x_open_display_with_name(display_name: POINTER): POINTER is
    external "C"
    alias "x_open_display_with_name"
    end;

  c_x_close_display(x11_disp: POINTER): POINTER is
    external "C"
    alias "x_close_display"
    end;

  c_x_flush(x11_disp: POINTER) is
    external "C"
    alias "x_flush"
    end;

  c_x_display_height(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_display_height"
    end;

  c_x_display_width(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_display_width"
    end;

  c_x_display_height_mm(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_display_height_mm"
    end;

  c_x_display_width_mm(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_display_width_mm"
    end;

  c_x_lock_display(x11_disp: POINTER) is
    external "C"
    alias "x_lock_display"
    end;

  c_x_unlock_display(x11_disp: POINTER) is
    external "C"
    alias "x_unlock_display"
    end;

  c_x_next_event(x11_disp: POINTER) is
    external "C"
    alias "x_next_event"
    end;

  c_x_last_event(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_last_event"
    end;

  c_x_is_last_event_window(x11_disp,x11_win: POINTER): BOOLEAN is
    external "C"
    alias "x_is_last_event_window"
    end;

  c_x_last_event_pointer_x_pos(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_last_event_pointer_x_pos"
    end;

  c_x_last_event_pointer_y_pos(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_last_event_pointer_y_pos"
    end;

  c_x_keyboard_event_to_string(x11_disp: POINTER): POINTER is
    external "C"
    alias "x_keyboard_event_to_string"
    end;

  c_x_keyboard_event_keysym(x11_disp: POINTER): INTEGER is
    external "C"
    alias "x_keyboard_event_keysym"
    end;

end -- X11_DISPLAY

