-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, May 2003
--

class X11_WINDOW

inherit
  X11_GLOBAL

creation
  make

feature

  make(disp: X11_DISPLAY) is
    require
      disp /= Void
    do
      display := disp;
        -- default attributes
      set_coordinates(0,0,10,10);
      set_border_width(5);
      set_border_color(display.black_color);
      set_background_color(display.white_color);
      disp.register_window(Current);
      created := false
    end;

  display: X11_DISPLAY;

feature -- creation attributes

  set_coordinates(w_x,w_y,w_width,w_heigth: INTEGER) is
    do
      creation_x := w_x;
      creation_y := w_y;
      creation_width := w_width;
      creation_height := w_heigth
    end;

  creation_x,creation_y: INTEGER; -- upper left corner relative to parent window
                                  -- reference (0,0) inside window (top left)

  creation_width,creation_height: INTEGER;

  set_border_width(bw: INTEGER) is
    do
      creation_border_width := bw
    end;

  creation_border_width: INTEGER;

  set_border_color(bc: X11_COLOR) is
    do
      border_color := bc
    end;

  border_color: X11_COLOR;

  set_background_color(bc: X11_COLOR) is
    do
      background_color := bc
    end;

  background_color: X11_COLOR;

feature

  cursor: X11_CURSOR;

  define_cursor(c: X11_CURSOR) is
    require
      not cursor_defined;
      c /= Void and then c.created
    do
      cursor := c;
      c_x_define_cursor(win_var,c.curs_var)
    end;

  cursor_defined: BOOLEAN is
    do
      Result := cursor /= Void
    end;

  define_cursor_from_shape(shape: INTEGER) is
    require
      not cursor_defined
    local
      c: X11_CURSOR
    do
      !!c.make(display);
      c.create_from_shape(shape);
      define_cursor(c)
    end;

  undefine_cursor is
    require
      cursor_defined
    do
      c_x_undefine_cursor(win_var);
      cursor := Void;
    end;

feature -- current window's attributes (fetched from actual window's state)

  origin_x: INTEGER is
    require
      created
    do
      Result := c_x_get_window_origin_x(win_var)
    end;

  origin_y: INTEGER is
    require
      created
    do
      Result := c_x_get_window_origin_y(win_var)
    end;

  width: INTEGER is
    require
      created
    do
      Result := c_x_get_window_origin_width(win_var)
    end;

  height: INTEGER is
    require
      created
    do
      Result := c_x_get_window_origin_height(win_var)
    end;

  border_width: INTEGER is
    require
      created
    do
      Result := c_x_get_window_origin_border_width(win_var)
    end;

  depth: INTEGER is
    require
      created
    do
      Result := c_x_get_window_origin_depth(win_var)
    end;

feature

  create_window(parent: X11_WINDOW) is
      -- root window if parent is Void
      -- name with "_window" to avoid interference with keyword create
    require
      not created
    do
      if parent = Void then
        win_var := c_x_create_simple_window(display.disp_var,
                                            creation_x,creation_y,
                                            creation_width,creation_height,
                                            creation_border_width,
                                            border_color.color_var,
                                            background_color.color_var)
      else
        win_var := c_x_create_simple_window_with_parent(display.disp_var,
                                            parent.win_var,
                                            creation_x,creation_y,
                                            creation_width,creation_height,
                                            creation_border_width,
                                            border_color.color_var,
                                            background_color.color_var)
      end;
        -- default initializations
      created := true;
      select_input(exposure_mask)
    end;

  destroy_window is
    require
      created
    do
      win_var := c_x_destroy_window(display.disp_var,win_var)
    end;

  created: BOOLEAN;

  map_raised is
    require
      created
    do
      c_x_map_raised(display.disp_var,win_var)
    end;

  map is
    require
      created
    do
      c_x_map_window(display.disp_var,win_var)
    end;

  unmap is
    require
      created
    do
      c_x_unmap_window(display.disp_var,win_var)
    end;

feature -- stacking order

  raise is
    require
      created
    do
      c_x_raise_window(display.disp_var,win_var)
    end;

  lower is
    require
      created
    do
      c_x_lower_window(display.disp_var,win_var)
    end;

feature -- input select

  input_event_mask: BIT 32;

  select_input(event_mask: BIT 32) is
    require
      created
    do
      input_event_mask := event_mask;
      c_x_select_input(win_var,input_event_mask.to_integer)
    end;

feature -- GC used in all output (one per-thread)

  graphic_context: X11_GRAPHIC_CONTEXT is
    require
      created
    do
      Result := internal_graphic_context.item;
      if Result = Void then -- default GC
        !!Result.create_new(Current);
        set_graphic_context(Result)
      end
    end;

  set_graphic_context(gc: X11_GRAPHIC_CONTEXT) is
    do
      internal_graphic_context.put(gc)
    end;

feature

  flush_on: BOOLEAN; -- flushs in every output X request

  enable_flush is
    do
      flush_on := true
    end;

  disable_flush is
    do
      flush_on := false
    end;

feature

  clear is
    require
      created
    do
      c_x_clear_window(display.disp_var,win_var);
      flush_if_requested
    end;

  clear_area_with_expose_events(px,py,rect_width,rect_height: INTEGER) is
    require
      created
    do
      c_x_clear_area(display.disp_var,win_var,px,py,rect_width,rect_height,1);
      flush_if_requested
    end;

  clear_area_without_expose_events(px,py,rect_width,rect_height: INTEGER) is
    require
      created
    do
      c_x_clear_area(display.disp_var,win_var,px,py,rect_width,rect_height,0);
      flush_if_requested
    end;

feature -- drawing  -  (0,0) -> topleft window point
        -- drawing attributes are located/defined in the Graphic Context

  draw_point(px,py:INTEGER) is
    require
      created
    do
      c_x_draw_point(win_var,graphic_context.gc_var,px,py);
      flush_if_requested
    end;

  add_point_to_draw(px,py:INTEGER) is
    require
      -- applies to 'draw_points', 'draw_lines' and 'fill_polygon'
      created
    do
      c_add_point_to_draw(px,py)
    end;

  clear_points is
    require
      created
    do
      c_clear_point_list
    end;

  drawing_mode: INTEGER is
    require
      -- default 'coord_mode_origin'
      -- applies to 'draw_points', 'draw_lines' and 'fill_polygon'
      created
    do
      Result := internal_drawing_mode.item;
      if Result = 0 then -- default initialization
        Result := coord_mode_origin;
        set_drawing_mode(Result)
      end
    ensure
      Result = coord_mode_origin or Result = coord_mode_previous
    end;

  set_drawing_mode(mode: INTEGER) is
    require
      created;
      mode = coord_mode_origin or mode = coord_mode_previous
    do
      internal_drawing_mode.put(mode)
    end;

  draw_points is
    require
      created
    do
      c_x_draw_points(win_var,graphic_context.gc_var,drawing_mode);
      flush_if_requested
    end;

  draw_lines is
    require
      created
    do
      c_x_draw_lines(win_var,graphic_context.gc_var,drawing_mode);
      flush_if_requested
    end;

  draw_line(x1,y1,x2,y2:INTEGER) is
    require
      created
    do
      c_x_draw_line(win_var,graphic_context.gc_var,x1,y1,x2,y2);
      flush_if_requested
    end;

  add_segment_to_draw(x1,y1,x2,y2:INTEGER) is
    require
      -- applies to 'draw_segments'
      created
    do
      c_add_segment_to_draw(x1,y1,x2,y2)
    end;

  clear_segments is
    require
      created
    do
      c_clear_segment_list
    end;

  draw_segments is
    require
      created
    do
      c_x_draw_segments(win_var,graphic_context.gc_var);
      flush_if_requested
    end;

feature -- drawing: arcs, circles and ellipses

  draw_arc(px,py,rect_width,rect_height,angle1,angle2: INTEGER) is
      -- (px,py): top left arc enclosing rectangle point
      -- rect_width: enclosing rectangle width
      -- rect_height: enclosing rectangle height
      -- angle1: angle in 1/64 degrees from horizontal line to the
      --         start of the arc
      -- angle2: angle in 1/64 degrees from angle1 until the end of the arc
      --
      -- Example full circle: draw_arc(x,y,side,side,0,64*360)
    require
      created;
      rect_width > 0;
      rect_height > 0;
    do
      c_x_draw_arc(win_var,graphic_context.gc_var,px,py,rect_width,rect_height,angle1,angle2);
      flush_if_requested
    end;

  draw_circle(px,py,radius: INTEGER) is
    require
      created;
      radius > 0;
    do
      draw_arc(px-radius,py-radius,2*radius,2*radius,0,64*360)
    end;

  add_arc_to_draw(px,py,rect_width,rect_height,angle1,angle2:INTEGER) is
    require
      -- applies to 'draw_arcs' and 'fill_arcs'
      created
    do
      c_add_arc_to_draw(px,py,rect_width,rect_height,angle1,angle2)
    end;

  clear_arcs is
    require
      created
    do
      c_clear_arc_list
    end;

  draw_arcs is
    require
      created
    do
      c_x_draw_arcs(win_var,graphic_context.gc_var);
      flush_if_requested
    end;

  fill_arc(px,py,rect_width,rect_height,angle1,angle2: INTEGER) is
    require
      created;
      rect_width > 0;
      rect_height > 0;
    do
      c_x_fill_arc(win_var,graphic_context.gc_var,px,py,rect_width,rect_height,angle1,angle2);
      flush_if_requested
    end;

  fill_circle(px,py,radius: INTEGER) is
    require
      created;
      radius > 0;
    do
      fill_arc(px-radius,py-radius,2*radius,2*radius,0,64*360)
    end;

  fill_arcs is
    require
      created
    do
      c_x_fill_arcs(win_var,graphic_context.gc_var);
      flush_if_requested
    end;

feature -- rectangles

  draw_rectangle(px,py,rect_width,rect_height: INTEGER) is
      -- (px,py): top left rectangle point
      -- rect_width: rectangle width
      -- rect_height: rectangle height
    require
      created;
      rect_width > 0;
      rect_height > 0;
    do
      c_x_draw_rectangle(win_var,graphic_context.gc_var,px,py,rect_width,rect_height);
      flush_if_requested
    end;

  add_rectangle_to_draw(px,py,rect_width,rect_height:INTEGER) is
    require
      -- applies to 'draw_rectangles' and 'fill_rectangles'
      created
    do
      c_add_rectangle_to_draw(px,py,rect_width,rect_height)
    end;

  clear_rectangles is
    require
      created
    do
      c_clear_rectangle_list
    end;

  draw_rectangles is
    require
      created
    do
      c_x_draw_rectangles(win_var,graphic_context.gc_var);
      flush_if_requested
    end;

  fill_rectangle(px,py,rect_width,rect_height: INTEGER) is
    require
      created;
      rect_width > 0;
      rect_height > 0;
    do
      c_x_fill_rectangle(win_var,graphic_context.gc_var,px,py,rect_width,rect_height);
      flush_if_requested
    end;

  fill_rectangles is
    require
      created
    do
      c_x_fill_rectangles(win_var,graphic_context.gc_var);
      flush_if_requested
    end;

feature -- polygons

  fill_polygon(shape: INTEGER) is
    require
      created;
      shape = complex or shape = nonconvex or shape = convex
    do
      c_x_fill_polygon(win_var,graphic_context.gc_var,shape,drawing_mode);
      flush_if_requested
    end;

feature -- text

  draw_string(px,py: INTEGER;str: STRING) is
      -- (px,py) relative to the origin of window
      --   px: leftmost string position
      --   py: baseline
    require
      created
    do
      c_x_draw_string(win_var,graphic_context.gc_var,px,py,str.to_external,str.count);
      flush_if_requested
    end;

  draw_image_string(px,py: INTEGER;str: STRING) is
      -- fill_style GC attribute ignored
    require
      created
    do
      c_x_draw_image_string(win_var,graphic_context.gc_var,px,py,str.to_external,str.count);
      flush_if_requested
    end;

feature {X11_DISPLAY,X11_WINDOW,X11_GRAPHIC_CONTEXT}

  win_var: POINTER;

feature {NONE}

  flush_if_requested is
    do
      if flush_on then
        display.flush
      end
    end;

  internal_graphic_context: THREAD_ATTRIBUTE[X11_GRAPHIC_CONTEXT];

  internal_drawing_mode: THREAD_ATTRIBUTE[INTEGER];

  c_x_create_simple_window(x11_disp: POINTER;w_x,w_y: INTEGER;
                           w_width,w_height: INTEGER;w_border_width:INTEGER;
                           w_border_color,w_background_color: POINTER): POINTER is
    external "C"
    alias "x_create_simple_window"
    end;

  c_x_create_simple_window_with_parent(x11_disp,x11_par:POINTER;w_x,w_y: INTEGER;
                           w_width,w_height: INTEGER;w_border_width:INTEGER;
                           w_border_color,w_background_color: POINTER): POINTER is
    external "C"
    alias "x_create_simple_window_with_parent"
    end;

  c_x_destroy_window(x11_disp,x11_win:POINTER): POINTER is
    external "C"
    alias "x_destroy_window"
    end;

  c_x_get_window_origin_x(x11_win:POINTER): INTEGER is
    external "C"
    alias "x_get_window_origin_x"
    end;

  c_x_get_window_origin_y(x11_win:POINTER): INTEGER is
    external "C"
    alias "x_get_window_origin_y"
    end;

  c_x_get_window_origin_width(x11_win:POINTER): INTEGER is
    external "C"
    alias "x_get_window_origin_width"
    end;

  c_x_get_window_origin_height(x11_win:POINTER): INTEGER is
    external "C"
    alias "x_get_window_origin_height"
    end;

  c_x_get_window_origin_border_width(x11_win:POINTER): INTEGER is
    external "C"
    alias "x_get_window_origin_border_width"
    end;

  c_x_get_window_origin_depth(x11_win:POINTER): INTEGER is
    external "C"
    alias "x_get_window_origin_depth"
    end;

  c_x_map_raised(x11_disp,x11_win:POINTER) is
    external "C"
    alias "x_map_raised"
    end;

  c_x_map_window(x11_disp,x11_win:POINTER) is
    external "C"
    alias "x_map_window"
    end;

  c_x_unmap_window(x11_disp,x11_win:POINTER) is
    external "C"
    alias "x_unmap_window"
    end;

  c_x_raise_window(x11_disp,x11_win:POINTER) is
    external "C"
    alias "x_raise_window"
    end;

  c_x_lower_window(x11_disp,x11_win: POINTER) is
    external "C"
    alias "x_raise_window"
    end;

  c_x_select_input(x11_win:POINTER;mask: INTEGER) is
    external "C"
    alias "x_select_input"
    end;

  c_x_clear_window(x11_disp,x11_win: POINTER) is
    external "C"
    alias "x_clear_window"
    end;

  c_x_clear_area(x11_disp,x11_win: POINTER;px,py,rect_width,rect_height,exposures: INTEGER) is
    external "C"
    alias "x_clear_area"
    end;

  c_x_draw_point(x11_win,x11_gc:POINTER;px,py: INTEGER) is
    external "C"
    alias "x_draw_point"
    end;

  c_add_point_to_draw(px,py: INTEGER) is
    external "C"
    alias "add_point_to_draw"
    end;

  c_clear_point_list is
    external "C"
    alias "clear_point_list"
    end;

  c_x_draw_points(x11_win,x11_gc:POINTER;mode: INTEGER) is
    external "C"
    alias "x_draw_points"
    end;

  c_x_draw_lines(x11_win,x11_gc:POINTER;mode: INTEGER) is
    external "C"
    alias "x_draw_lines"
    end;

  c_x_draw_line(x11_win,x11_gc:POINTER;x1,y1,x2,y2: INTEGER) is
    external "C"
    alias "x_draw_line"
    end;

  c_add_segment_to_draw(x1,y1,x2,y2: INTEGER) is
    external "C"
    alias "add_segment_to_draw"
    end;

  c_clear_segment_list is
    external "C"
    alias "clear_segment_list"
    end;

  c_x_draw_segments(x11_win,x11_gc:POINTER) is
    external "C"
    alias "x_draw_segments"
    end;

  c_x_draw_arc(x11_win,x11_gc:POINTER;px,py,rect_width,rect_height,angle1,angle2: INTEGER) is
    external "C"
    alias "x_draw_arc"
    end;

  c_add_arc_to_draw(px,py,rect_width,rect_height,angle1,angle2: INTEGER) is
    external "C"
    alias "add_arc_to_draw"
    end;

  c_clear_arc_list is
    external "C"
    alias "clear_arc_list"
    end;

  c_x_draw_arcs(x11_win,x11_gc:POINTER) is
    external "C"
    alias "x_draw_arcs"
    end;

  c_x_fill_arc(x11_win,x11_gc:POINTER;px,py,rect_width,rect_height,angle1,angle2: INTEGER) is
    external "C"
    alias "x_fill_arc"
    end;

  c_x_fill_arcs(x11_win,x11_gc:POINTER) is
    external "C"
    alias "x_fill_arcs"
    end;

  c_x_draw_rectangle(x11_win,x11_gc:POINTER;px,py,rect_width,rect_height: INTEGER) is
    external "C"
    alias "x_draw_rectangle"
    end;

  c_add_rectangle_to_draw(px,py,rect_width,rect_height: INTEGER) is
    external "C"
    alias "add_rectangle_to_draw"
    end;

  c_clear_rectangle_list is
    external "C"
    alias "clear_rectangle_list"
    end;

  c_x_draw_rectangles(x11_win,x11_gc:POINTER) is
    external "C"
    alias "x_draw_rectangles"
    end;

  c_x_fill_rectangle(x11_win,x11_gc:POINTER;px,py,rect_width,rect_height: INTEGER) is
    external "C"
    alias "x_fill_rectangle"
    end;

  c_x_fill_rectangles(x11_win,x11_gc:POINTER) is
    external "C"
    alias "x_fill_rectangles"
    end;

  c_x_fill_polygon(x11_win,x11_gc:POINTER;shape,mode: INTEGER) is
    external "C"
    alias "x_fill_polygon"
    end;

  c_x_draw_string(x11_win,x11_gc:POINTER;px,py: INTEGER;str: POINTER;length: INTEGER) is
    external "C"
    alias "x_draw_string"
    end;

  c_x_draw_image_string(x11_win,x11_gc:POINTER;px,py: INTEGER;str: POINTER;length: INTEGER) is
    external "C"
    alias "x_draw_image_string"
    end;

  c_x_define_cursor(x11_win,x11_cur:POINTER) is
    external "C"
    alias "x_define_cursor"
    end;

  c_x_undefine_cursor(x11_win:POINTER) is
    external "C"
    alias "x_undefine_cursor"
    end;

end -- X11_WINDOW

