/* chr_term.c
** This  file  is   free  software,  which   comes  along  with   thread-safe
** SmallEiffel. This software  is distributed  in the  hope that  it will  be
** useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
** you want, provided this  header is kept unaltered,  and a notification  of
** the changes is  added. You  are allowed to  redistribute it  and sell  it,
** alone or as a part of another product.
**
** (C) Miguel Oliveira e Silva, August 2001
*/

#include "chr_term.h"
#include <curses.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>

/* from eiffel defined key codes */

#define key_minimum_event_code 0
#define key_maximum_event_code 1150

#define key_none            0
#define key_tab             9
#define key_return          10
#define key_escape          27
#define key_up              1000
#define key_down            1001
#define key_left            1002
#define key_right           1003
#define key_home            1004
#define key_end             1005
#define key_insert          1006
#define key_delete          1007
#define key_page_up         1008
#define key_page_down       1009
#define key_back_space      1010
#define key_shift_up        1011
#define key_shift_down      1012
#define key_shift_left      1013
#define key_shift_right     1014
#define key_shift_home      1015
#define key_shift_end       1016
#define key_shift_page_up   1017
#define key_shift_page_down 1018
#define key_f1              1101
#define key_f2              1102
#define key_f3              1103
#define key_f4              1104
#define key_f5              1105
#define key_f6              1106
#define key_f7              1107
#define key_f8              1108
#define key_f9              1109
#define key_f10             1110
#define key_f11             1111
#define key_f12             1112

#define mouse_minimum_event_code 1200
#define mouse_maximum_event_code 1250

#define mouse_button1_press          1200
#define mouse_button1_released       1201
#define mouse_button1_clicked        1202
#define mouse_button1_double_clicked 1203
#define mouse_button1_triple_clicked 1204
#define mouse_button2_press          1205
#define mouse_button2_released       1206
#define mouse_button2_clicked        1207
#define mouse_button2_double_clicked 1208
#define mouse_button2_triple_clicked 1209
#define mouse_button3_press          1210
#define mouse_button3_released       1211
#define mouse_button3_clicked        1212
#define mouse_button3_double_clicked 1213
#define mouse_button3_triple_clicked 1214
#define mouse_movement               1215


/* from eiffel defined event codes */

#define button1_press_mask          0x00001
#define button1_released_mask       0x00002
#define button1_clicked_mask        0x00004
#define button1_double_clicked_mask 0x00008
#define button1_triple_clicked_mask 0x00010
#define button2_press_mask          0x00020
#define button2_released_mask       0x00040
#define button2_clicked_mask        0x00080
#define button2_double_clicked_mask 0x00100
#define button2_triple_clicked_mask 0x00200
#define button3_press_mask          0x00400
#define button3_released_mask       0x00800
#define button3_clicked_mask        0x01000
#define button3_double_clicked_mask 0x02000
#define button3_triple_clicked_mask 0x04000
#define mouse_movement_mask         0x08000
#define all_mouse_events_mask       0x0FFFF


static int n_lines = 0;
static int n_columns = 0;

static void finish(int sig)
{
  end_display();
}

static int init = 0;

int initialized(void)
{
  return init;
}

void init_display(void)
{
  if (init == 0)
  {
    init = 1;
    signal(SIGINT, finish);
    initscr();
    noecho();
    cbreak();
    leaveok(stdscr,FALSE);
    keypad(stdscr,TRUE);
    meta(stdscr,TRUE);
    n_lines = LINES;
    n_columns = COLS;
    if (has_colors())
    {
      start_color();
      init_pair(1, COLOR_RED,     COLOR_BLACK);
      init_pair(2, COLOR_GREEN,   COLOR_BLACK);
      init_pair(3, COLOR_YELLOW,  COLOR_BLACK);
      init_pair(4, COLOR_BLUE,    COLOR_BLACK);
      init_pair(5, COLOR_CYAN,    COLOR_BLACK);
      init_pair(6, COLOR_MAGENTA, COLOR_BLACK);
      init_pair(7, COLOR_WHITE,   COLOR_BLACK);
    }
  }
}

void end_display(void)
{
  if (init == 1)
  {
    init = 0;
    endwin();
  }
}

int number_of_lines(void)
{
  return n_lines;
}

int number_of_columns(void)
{
  return n_columns;
}

static int cursor_visible = 0;

int is_cursor_visible(void)
{
  return cursor_visible;
}

void hide_cursor(void)
{
  cursor_visible = 0;
  curs_set(0);
}

void show_cursor(void)
{
  cursor_visible = 1;
  curs_set(1);
}

static int mouse_enabled = 0;

int is_mouse_usable(void)
{
#ifdef NCURSES_MOUSE_VERSION
  static int first=1;
  static int result = 0;

  if (first)
  {
    mmask_t m;

    first = 0;
    if (mousemask(ALL_MOUSE_EVENTS,&m))
    {
      result = 1;
      mousemask(m,NULL);
    }
  }

  return result;
#else
  return 0
#endif /* NCURSES_MOUSE_VERSION */
}

void enable_mouse(void)
{
  mouse_enabled = 1;
}

void disable_mouse(void)
{
  mouse_enabled = 0;
#ifdef NCURSES_MOUSE_VERSION
  mousemask(0,NULL);
#endif /* NCURSES_MOUSE_VERSION */
}

int is_mouse_enabled(void)
{
  return mouse_enabled;
}

#ifdef NCURSES_MOUSE_VERSION
static mmask_t translate_mask(unsigned mask)
{
  mmask_t result;

  result = 0UL;
  if (mask & button1_press_mask)
    result |= BUTTON1_PRESSED;
  if (mask & button1_released_mask)
    result |= BUTTON1_RELEASED;
  if (mask & button1_clicked_mask)
    result |= BUTTON1_CLICKED;
  if (mask & button1_double_clicked_mask)
    result |= BUTTON1_DOUBLE_CLICKED;
  if (mask & button1_triple_clicked_mask)
    result |= BUTTON1_TRIPLE_CLICKED;
  if (mask & button2_press_mask)
    result |= BUTTON2_PRESSED;
  if (mask & button2_released_mask)
    result |= BUTTON2_RELEASED;
  if (mask & button2_clicked_mask)
    result |= BUTTON2_CLICKED;
  if (mask & button2_double_clicked_mask)
    result |= BUTTON2_DOUBLE_CLICKED;
  if (mask & button2_triple_clicked_mask)
    result |= BUTTON2_TRIPLE_CLICKED;
  if (mask & button3_press_mask)
    result |= BUTTON3_PRESSED;
  if (mask & button3_released_mask)
    result |= BUTTON3_RELEASED;
  if (mask & button3_clicked_mask)
    result |= BUTTON3_CLICKED;
  if (mask & button3_double_clicked_mask)
    result |= BUTTON3_DOUBLE_CLICKED;
  if (mask & button3_triple_clicked_mask)
    result |= BUTTON3_TRIPLE_CLICKED;
  if (mask & mouse_movement_mask)
    result |= REPORT_MOUSE_POSITION;

  return result;
}
#endif /* NCURSES_MOUSE_VERSION */

static int mouse_event_mask = 0;

int get_mouse_event_mask(void)
{
  return mouse_event_mask;
}

void set_mouse_event_mask(int m)
{
  mouse_event_mask = m;
#ifdef NCURSES_MOUSE_VERSION
  mousemask(translate_mask((unsigned)m),NULL);
#endif /* NCURSES_MOUSE_VERSION */
}

int is_mouse_event_mask_usable(int m)
{
#ifdef NCURSES_MOUSE_VERSION
  mmask_t mask,oldmask;
  int result;

  mask = translate_mask((unsigned)m);

  if (mousemask(mask,&oldmask) == mask)
    result = 1;
  else
    result = 0;
  mousemask(oldmask,NULL);

  return result;
#else
  return 0;
#endif /* NCURSES_MOUSE_VERSION */
}


static int keyboard_enabled = 0;

void enable_keyboard(void)
{
  keyboard_enabled = 1;
}

void disable_keyboard(void)
{
  keyboard_enabled = 0;
}

int is_keyboard_enabled(void)
{
  return keyboard_enabled;
}


static int mouse_x = 1;
static int mouse_y = 1;

int last_mouse_x(void)
{
  return mouse_x;
}

int last_mouse_y(void)
{
  return mouse_y;
}

static int lastevent = 0;

int last_event(void)
{
  return lastevent;
}

void set_timeout(int ts)
{
  timeout(ts);
}

void unset_timeout(void)
{
  timeout(0);
}

void get_event(void)
{
  int c,o;
  int wait_valid_event;
  MEVENT mev;

  wait_valid_event = 1;
  while(wait_valid_event)
  {
    c = getch();
#ifdef NCURSES_MOUSE_VERSION
    if (c == KEY_MOUSE)
    {
      getmouse(&mev);
      if (mouse_enabled)
      {
        switch(mev.bstate)
        {
          case BUTTON1_PRESSED:         c = mouse_button1_press; break;
          case BUTTON1_RELEASED:        c = mouse_button1_released; break;
          case BUTTON1_CLICKED:         c = mouse_button1_clicked; break;
          case BUTTON1_DOUBLE_CLICKED:  c = mouse_button1_double_clicked; break;
          case BUTTON1_TRIPLE_CLICKED:  c = mouse_button1_triple_clicked; break;
          case BUTTON2_PRESSED:         c = mouse_button2_press; break;
          case BUTTON2_RELEASED:        c = mouse_button2_released; break;
          case BUTTON2_CLICKED:         c = mouse_button2_clicked; break;
          case BUTTON2_DOUBLE_CLICKED:  c = mouse_button2_double_clicked; break;
          case BUTTON2_TRIPLE_CLICKED:  c = mouse_button2_triple_clicked; break;
          case BUTTON3_PRESSED:         c = mouse_button3_press; break;
          case BUTTON3_RELEASED:        c = mouse_button3_released; break;
          case BUTTON3_CLICKED:         c = mouse_button3_clicked; break;
          case BUTTON3_DOUBLE_CLICKED:  c = mouse_button3_double_clicked; break;
          case BUTTON3_TRIPLE_CLICKED:  c = mouse_button3_triple_clicked; break;
          case REPORT_MOUSE_POSITION:   c = mouse_movement; break;
        }
        mouse_x = mev.x+1;
        mouse_y = mev.y+1;
        wait_valid_event = 0;
      }
    }
    else
#endif /* NCURSES_MOUSE_VERSION */
    if (keyboard_enabled)
    {
      switch(c)
      {
        case ERR:           c = key_none;       break;
        case KEY_UP:        c = key_up;         break;
        case KEY_DOWN:      c = key_down;       break;
        case KEY_LEFT:      c = key_left;       break;
        case KEY_RIGHT:     c = key_right;      break;
        case KEY_HOME:      c = key_home;       break;
        case KEY_END:       c = key_end;        break;
        case KEY_IC:        c = key_insert;     break;
        case KEY_DC:        c = key_delete;     break;
        case KEY_PPAGE:     c = key_page_up;    break;
        case KEY_NPAGE:     c = key_page_down;  break;
        case KEY_BACKSPACE: c = key_back_space; break;
/*
        case KEY_SPREVIOUS: c = key_shift_up;         break;
        case KEY_SNEXT:     c = key_shift_down;       break;
*/
        case KEY_SLEFT:     c = key_shift_left;       break;
        case KEY_SRIGHT:    c = key_shift_right;      break;
        case KEY_SHOME:     c = key_shift_home;       break;
        case KEY_SEND:      c = key_shift_end;        break;
/*
        case KEY_SPPAGE:    c = key_shift_page_up;    break;
        case KEY_SNPAGE:    c = key_shift_page_down;  break;
*/
        default:
          o = c - KEY_F(1);
          if(o >= 0 && o < 12)
            c = key_f1 + o;
          break;
      }
      if (c >= key_minimum_event_code && c <= key_maximum_event_code)
        wait_valid_event = 0;
    }
  }
  lastevent = c;
  if (!cursor_visible)
    hide_cursor();
}

static int cursor_lin = 1;
static int cursor_col = 1;

int last_cursor_lin(void)
{
  return cursor_lin;
}

int last_cursor_col(void)
{
  return cursor_col;
}

void set_cursor(int lin,int col)
{
  move(lin-1,col-1);
  refresh();
}

void get_cursor(void)
{
  getsyx(cursor_lin,cursor_col);
  cursor_lin++;
  cursor_col++;
}

void do_beep(void)
{
  beep();
}

void clear_screen(void)
{
  erase();
  refresh();
}

static int color = 0;
static int attr = A_NORMAL;

void print_char(char c)
{
/*
  static char t[2] = " ";

  t[0] = c;
  t[1] = '\0';
*/
  addch((chtype)c | attr | color);
  refresh();
}

void set_normal_text(void)
{
  attr = A_NORMAL;	
}

void set_highlight(void)
{
  attr |= A_STANDOUT;	
}

void reset_highlight(void)
{
  attr &= (~A_STANDOUT);	
}

void set_underline(void)
{
  attr |= A_UNDERLINE;	
}

void reset_underline(void)
{
  attr &= (~A_UNDERLINE);	
}

void set_reverse(void)
{
  attr |= A_REVERSE;	
}

void reset_reverse(void)
{
  attr &= (~A_REVERSE);	
}

void set_blinking(void)
{
  attr |= A_BLINK;	
}

void reset_blinking(void)
{
  attr &= (~A_BLINK);	
}

void set_dim(void)
{
  attr |= A_DIM;	
}

void reset_dim(void)
{
  attr &= (~A_DIM);	
}

void set_bold(void)
{
  attr |= A_BOLD;	
}

void reset_bold(void)
{
  attr &= (~A_BOLD);	
}

void set_black_color(void)
{
  color = COLOR_PAIR(COLOR_BLACK);	
}

void set_red_color(void)
{
  color = COLOR_PAIR(1);	
}

void set_green_color(void)
{
  color = COLOR_PAIR(2);	
}

void set_yellow_color(void)
{
  color = COLOR_PAIR(3);	
}

void set_blue_color(void)
{
  color = COLOR_PAIR(4);	
}

void set_cyan_color(void)
{
  color = COLOR_PAIR(5);	
}

void set_magenta_color(void)
{
  color = COLOR_PAIR(6);	
}

void set_white_color(void)
{
  color = COLOR_PAIR(7);	
}

static pthread_mutex_t term_hand_mutex = PTHREAD_MUTEX_INITIALIZER;

void term_hand_lock(void)
{
  pthread_mutex_lock(&term_hand_mutex);
}

void term_hand_unlock(void)
{
  pthread_mutex_unlock(&term_hand_mutex);
}

