-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, August 2001
--
-- Thread-safe interface to TERMINAL_HANDLER
--
-- All thread-unsafe features are renamed to 'unsafe_*'
--

class THREAD_TERMINAL_HANDLER

inherit
  TERMINAL_HANDLER
    rename
      initialize as unsafe_initialize,
      terminate as unsafe_terminate,
      number_of_lines as unsafe_number_of_lines,
      number_of_columns as unsafe_number_of_columns,
      cursor_visible as unsafe_cursor_visible,
      hide_cursor as unsafe_hide_cursor,
      show_cursor as unsafe_show_cursor,
      mouse_usable as unsafe_mouse_usable,
      enable_mouse as unsafe_enable_mouse,
      disable_mouse as unsafe_disable_mouse,
      mouse_enabled as unsafe_mouse_enabled,
      valid_mouse_event_mask as safe_valid_mouse_event_mask,
      mouse_event_mask_usable as unsafe_mouse_event_mask_usable,
      mouse_event_mask as unsafe_mouse_event_mask,
      set_mouse_event_mask as unsafe_set_mouse_event_mask,
      enable_keyboard as unsafe_enable_keyboard,
      disable_keyboard as unsafe_disable_keyboard,
      keyboard_enabled as unsafe_keyboard_enabled,
      set_timeout as unsafe_set_timeout,
      unset_timeout as unsafe_unset_timeout,
      fetch_event as unsafe_fetch_event,
      last_event as unsafe_last_event,
      keyboard_event as unsafe_keyboard_event,
      mouse_event as unsafe_mouse_event,
      mouse_line as unsafe_mouse_line,
      mouse_column as unsafe_mouse_column,
      set_cursor_position as unsafe_set_cursor_position,
      get_cursor_position as unsafe_get_cursor_position,
      cursor_line as unsafe_cursor_line,
      cursor_column as unsafe_cursor_column,
      beep as unsafe_beep,
      clear_screen as unsafe_clear_screen,
      new_line as unsafe_new_line,
      write_char as unsafe_write_char,
      write_char_n_times as unsafe_write_char_n_times,
      write as unsafe_write,
      write_bounded as unsafe_write_bounded,
      writeln as unsafe_writeln,
      write_char_at as unsafe_write_char_at,
      write_at as unsafe_write_at,
      write_bounded_at as unsafe_write_bounded_at,
      writeln_at as unsafe_writeln_at,
      set_text_attributes as unsafe_set_text_attributes,
      set_normal_text as unsafe_set_normal_text,
      set_highlight as unsafe_set_highlight,
      reset_highlight as unsafe_reset_highlight,
      set_underline as unsafe_set_underline,
      reset_underline as unsafe_reset_underline,
      set_reverse as unsafe_set_reverse,
      reset_reverse as unsafe_reset_reverse,
      set_blinking as unsafe_set_blinking,
      reset_blinking as unsafe_reset_blinking,
      set_dim as unsafe_set_dim,
      reset_dim as unsafe_reset_dim,
      set_bold as unsafe_set_bold,
      reset_bold as unsafe_reset_bold,
      set_text_color as unsafe_set_text_color,
      set_black_color as unsafe_set_black_color,
      set_red_color as unsafe_set_red_color,
      set_green_color as unsafe_set_green_color,
      set_yellow_color as unsafe_set_yellow_color,
      set_blue_color as unsafe_set_blue_color,
      set_cyan_color as unsafe_set_cyan_color,
      set_magenta_color as unsafe_set_magenta_color,
      set_white_color as unsafe_set_white_color,
      edit_line_text as unsafe_edit_line_text
    end

feature

  initialize is
      -- global initialization on 'unsafe_initialize' and per-thread
      --  initialization on cursor position and text attributes and color!
    once
      lock;
      unsafe_initialize;
      unlock;
      get_cursor_position;
      set_text_attributes(normal_text_id);
      set_text_color(default_color_id);
    end;

  terminate is
    once
      lock;
      unsafe_terminate;
      unlock
    end;

feature -- These features are used to serialize TERMINAL_HANDLER access
        -- It can also be used for a larger-grained serialization of
        --  unsafe features

  lock is
    external "C"
    alias "term_hand_lock"
    end;

  unlock is
    external "C"
    alias "term_hand_unlock"
    end;

feature

  number_of_lines: INTEGER is
    do
      lock;
      Result := unsafe_number_of_lines;
      unlock
    end;
  
  number_of_columns: INTEGER is
    do
      lock;
      Result := unsafe_number_of_columns;
      unlock
    end;

feature

  cursor_visible: BOOLEAN is
    do
      lock;
      Result := unsafe_cursor_visible;
      unlock
    end;

  hide_cursor is
    do
      lock;
      unsafe_hide_cursor;
      unlock
    end;
  
  show_cursor is
    do
      lock;
      unsafe_show_cursor;
      unlock
    end;

feature -- mouse handling

  mouse_usable: BOOLEAN is
    do
      lock;
      Result := unsafe_mouse_usable;
      unlock
    end;

  enable_mouse is
    do
      lock;
      unsafe_enable_mouse;
      unlock
    end;

  disable_mouse is
    do
      lock;
      unsafe_disable_mouse;
      unlock
    end;

  mouse_enabled: BOOLEAN is
    do
      lock;
      Result := unsafe_mouse_enabled;
      unlock
    end;

  valid_mouse_event_mask(mask: BIT 24): BOOLEAN is
    do
      Result := safe_valid_mouse_event_mask(mask)
    end;

  mouse_event_mask_usable(mask: BIT 24): BOOLEAN is
    do
      lock;
      Result := unsafe_mouse_event_mask_usable(mask);
      unlock
    end;

  mouse_event_mask: BIT 24 is
    do
      lock;
      Result := unsafe_mouse_event_mask;
      unlock
    end;

  set_mouse_event_mask(mask: BIT 24) is
    require
      valid_mouse_event_mask(mask)
    do
      lock;
      unsafe_set_mouse_event_mask(mask);
      unlock
    end;

feature -- keyboard handling

  enable_keyboard is
    do
      lock;
      unsafe_enable_keyboard;
      unlock
    end;

  disable_keyboard is
    do
      lock;
      unsafe_disable_keyboard;
      unlock
    end;

  keyboard_enabled: BOOLEAN is
    do
      lock;
      Result := unsafe_keyboard_enabled;
      unlock
    end;

feature

  set_timeout(tenths_sec: INTEGER) is
      -- maximum wait delay in 'fetch_event'
      -- on timeout, 'key_none' is returned
      -- from 0.1 until 25.5 seconds (0.1 seconds step)
    require
      tenths_sec >= 1 and tenths_sec <= 255
    do
      lock;
      unsafe_set_timeout(tenths_sec);
      unlock
    end
  
  unset_timeout is
      -- default behavior
      -- no timeout in 'fetch_event'
    do
      lock;
      unsafe_unset_timeout;
      unlock
    end

feature

  fetch_event is
      -- atomicty gets all event information and makes a per-thread storage.
    do
      lock;
      unsafe_fetch_event;
      last_event_ref.set_item(unsafe_last_event);
      mouse_line_ref.set_item(unsafe_mouse_line);
      mouse_column_ref.set_item(unsafe_mouse_column);
      unlock
    end;

  last_event: INTEGER is
    do
      Result := last_event_ref.item
    end;

  keyboard_event: BOOLEAN is
      -- is last thread fetched event a keyboard event?
    do
      Result := (last_event_ref.item >= key_minimum_event_code) and
                (last_event_ref.item <= key_maximum_event_code)
    end;

  mouse_event: BOOLEAN is
      -- is last thread fetched event a mouse event?
    do
      Result := (last_event_ref.item >= mouse_minimum_event_code) and
                (last_event_ref.item <= mouse_maximum_event_code)
    end;

  mouse_line: INTEGER is
    -- last (thread) mouse event line position [1..number_of_lines]
    -- [1,1]: top left terminal position
    do
      Result := mouse_line_ref.item
    end;

  mouse_column: INTEGER is
    -- last (thread) mouse event column position [1..number_of_columns]
    do
      Result := mouse_column_ref.item
    end;

feature -- cursor

  set_cursor_position(lin,col: INTEGER) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      lock;
      unsafe_set_cursor_position(lin,col);
      cursor_line_ref.set_item(lin);
      cursor_column_ref.set_item(col);
      unlock
    end;

  get_cursor_position is
    do
      lock;
      thread_get_cursor_position;
      unlock
    end;

  cursor_line: INTEGER is
    -- last (thread) 'get_cursor_position' cursor line position
    --   [1..number_of_lines]
    -- [1,1]: top left terminal position
    do
      Result := cursor_line_ref.item
    end;

  cursor_column: INTEGER is
    -- last (thread) 'get_cursor_position' cursor column position
    --   [1..number_of_columns]
    do
      Result := cursor_column_ref.item
    end;

feature -- output handling

  beep is
    do
      lock;
      unsafe_beep;
      unlock
    end;

  clear_screen is
    do
      lock;
      unsafe_clear_screen;
      unlock
    end;

  new_line is
    do
      lock;
      unsafe_set_cursor_position(cursor_line,cursor_column);
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_new_line;
      thread_get_cursor_position;
      unlock
    end;

  write_char(c: CHARACTER) is
    do
      lock;
      unsafe_set_cursor_position(cursor_line,cursor_column);
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write_char(c);
      thread_get_cursor_position;
      unlock
    end;

  write_char_n_times(c: CHARACTER;n: INTEGER) is
    do
      lock;
      unsafe_set_cursor_position(cursor_line,cursor_column);
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write_char_n_times(c,n);
      thread_get_cursor_position;
      unlock
    end;

  write(txt: STRING) is
    do
      lock;
      unsafe_set_cursor_position(cursor_line,cursor_column);
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write(txt);
      thread_get_cursor_position;
      unlock
    end;

  write_bounded(txt: STRING;max_chars: INTEGER) is
    do
      lock;
      unsafe_set_cursor_position(cursor_line,cursor_column);
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write_bounded(txt,max_chars);
      thread_get_cursor_position;
      unlock
    end;

  writeln(txt: STRING) is
    do
      lock;
      unsafe_set_cursor_position(cursor_line,cursor_column);
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_writeln(txt);
      thread_get_cursor_position;
      unlock
    end;

  write_char_at(lin,col: INTEGER;c: CHARACTER) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      lock;
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write_char_at(lin,col,c);
      thread_get_cursor_position;
      unlock
    end;

  write_at(lin,col: INTEGER;txt: STRING) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      lock;
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write_at(lin,col,txt);
      thread_get_cursor_position;
      unlock
    end;

  write_bounded_at(lin,col: INTEGER;txt: STRING;max_chars: INTEGER) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      lock;
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_write_bounded_at(lin,col,txt,max_chars);
      thread_get_cursor_position;
      unlock
    end;

  writeln_at(lin,col: INTEGER;txt: STRING) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      lock;
      unsafe_set_text_attributes(text_attributes);
      unsafe_set_text_color(text_color);
      unsafe_writeln_at(lin,col,txt);
      thread_get_cursor_position;
      unlock
    end;

feature -- write properties

  set_text_attributes(a: BIT 24) is
    do
      lock;
      unsafe_set_text_attributes(a);
      unlock
    end;

  set_normal_text is
      -- Normal display (no highlight)
    do
      lock;
      unsafe_set_normal_text;
      unlock
    end;

  set_highlight is
      -- Best highlighting mode of the terminal
    do
      lock;
      unsafe_set_highlight;
      unlock
    end;

  reset_highlight is
    do
      lock;
      unsafe_reset_highlight;
      unlock
    end;

  set_underline is
    do
      lock;
      unsafe_set_underline;
      unlock
    end;

  reset_underline is
    do
      lock;
      unsafe_reset_underline;
      unlock
    end;

  set_reverse is
    do
      lock;
      unsafe_set_reverse;
      unlock
    end;

  reset_reverse is
    do
      lock;
      unsafe_reset_reverse;
      unlock
    end;

  set_blinking is
    do
      lock;
      unsafe_set_blinking;
      unlock
    end;

  reset_blinking is
    do
      lock;
      unsafe_reset_blinking;
      unlock
    end;

  set_dim is
      -- Half bright
    do
      lock;
      unsafe_set_dim;
      unlock
    end;

  reset_dim is
    do
      lock;
      unsafe_reset_dim;
      unlock
    end;

  set_bold is
      -- Extra bright
    do
      lock;
      unsafe_set_bold;
      unlock
    end;

  reset_bold is
    do
      lock;
      unsafe_reset_bold;
      unlock
    end;

feature -- color

  set_text_color(c: INTEGER) is
    do
      lock;
      unsafe_set_text_color(c);
      unlock
    end;

  set_black_color is
    do
      lock;
      unsafe_set_black_color;
      unlock
    end;

  set_red_color is
    do
      lock;
      unsafe_set_red_color;
      unlock
    end;

  set_green_color is
    do
      lock;
      unsafe_set_green_color;
      unlock
    end;

  set_yellow_color is
    do
      lock;
      unsafe_set_yellow_color;
      unlock
    end;

  set_blue_color is
    do
      lock;
      unsafe_set_blue_color;
      unlock
    end;

  set_cyan_color is
    do
      lock;
      unsafe_set_cyan_color;
      unlock
    end;

  set_magenta_color is
    do
      lock;
      unsafe_set_magenta_color;
      unlock
    end;

  set_white_color is
    do
      lock;
      unsafe_set_white_color;
      unlock
    end;

feature -- advanced features

  edit_line_text(lin,col,width: INTEGER; text: STRING;
                 fixed: INTEGER; cursor_at_end: BOOLEAN): BOOLEAN is
      -- true result if text changed (this is bad because it does not obey CQS).      -- (lin,col): leftmost position of edit window
      -- width: width edit window
      -- text: string to edit
      -- fixed: number of initial 'text' characters to remain unchangeable
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
      width > 0 and width <= number_of_columns - col + 1;
      text /= Void and then (fixed >= 0 and fixed <= text.count)
    do
      lock;
      Result := unsafe_edit_line_text(lin,col,width,text,fixed,cursor_at_end);
      unlock
    end;

feature {NONE}

  thread_get_cursor_position is
    do
      unsafe_get_cursor_position;
      cursor_line_ref.set_item(unsafe_cursor_line);
      cursor_column_ref.set_item(unsafe_cursor_column);
    end;

  last_event_ref: INTEGER_REF is
      -- one per thread!
    once
      !!Result
    end;

  mouse_line_ref: INTEGER_REF is
      -- one per thread!
    once
      !!Result
    end;

  mouse_column_ref: INTEGER_REF is
      -- one per thread!
    once
      !!Result
    end;

  cursor_line_ref: INTEGER_REF is
      -- one per thread!
    once
      !!Result
    end;

  cursor_column_ref: INTEGER_REF is
      -- one per thread!
    once
      !!Result
    end;

end -- THREAD_TERMINAL_HANDLER

