--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--
-- Changed to new global once in August 2003
--

class PARTICLE

inherit
  THREAD;
  THREAD_TERMINAL_HANDLER

creation
  make

feature {PARTICLE}

  main is
    do
      update_barrier.sign_on;
      initialize;
      hide_cursor;
      set_text_color(colors @ (random_integer(colors.count-1)+1));
      set_bold;
      last_line := 0;
      last_column := 0;
      convert_to_screen_coordinates;
      timer.reset_lost_periods;
      from
      until
        finish
      loop
        draw;
        timer.wait;
        estimating_advance_time_values;
        update_barrier.wait;
        advance_time;
        convert_to_screen_coordinates;
      end;
      clean_up;
      terminate;
      update_barrier.sign_off;
      particle_list.remove_if_present(Current)
    end;

feature -- [global] external gravity

  fetch_external_acceleration(x,y: DOUBLE;a: VECTOR) is
    do
      if x >= 1.0 and x <= number_of_columns.to_double then
        a.set_x(0.0)
      else
        if x < 1.0 then
          a.set_x((1.0 - x))
        else
          a.set_x((number_of_columns.to_double - x))
        end
      end;
      if y >= 1.0 and y/2.0 <= number_of_lines.to_double then
        a.set_y(0.0)
      else
        if y < 1.0 then
          a.set_y((1.0 - y))
        else
          a.set_y((number_of_lines.to_double - y/2.0))
        end
      end
    end;

feature -- screen output

  colors: ARRAY[INTEGER] is
    once {"process"}
      Result := <<red_color_id,green_color_id,
                  yellow_color_id,blue_color_id,
                  cyan_color_id,magenta_color_id>>;
    end;

  line,column: INTEGER;

  last_line,last_column: INTEGER;

  convert_to_screen_coordinates is
      -- 1 meter = 0.5 line = 1 column
    do
      read_lock;
      line := (position.y / 2.0).rounded;
      column := position.x.rounded;
      read_unlock;
    end;

  draw is
    do
      if line /= last_line or column /= last_column then
        if last_line >= 1 and last_line <= number_of_lines and
           last_column >= 1 and last_column <= number_of_columns then
          write_char_at(last_line,last_column,' ')
        end;
        if line >= 1 and line <= number_of_lines and
           column >= 1 and column <= number_of_columns then
          write_char_at(line,column,shape)
        end;
        last_line := line;
        last_column := column;
      end;
    end;

  clean_up is
    do
      if line >= 1 and line <= number_of_lines and
         column >= 1 and column <= number_of_columns then
        write_char_at(line,column,' ')
      end;
      if last_line >= 1 and last_line <= number_of_lines and
           last_column >= 1 and last_column <= number_of_columns then
          write_char_at(last_line,last_column,' ')
      end;
    end;

feature

  make(m: DOUBLE;p0,v0: VECTOR;ti: DOUBLE;tmr: TIMER;ub: THREAD_BARRIER;pl: PARTICLE_LIST) is
    do
      mass := m;
      set_shape;
      !!position;
      position.set(p0);
      !!velocity;
      velocity.set(v0);
      !!acceleration;
      acceleration.set_zero;
      time_increment := ti;
      timer := tmr;
      update_barrier := ub;
      particle_list := pl;
      !!new_position;
      !!new_velocity;
      is_finish := false;
      !!rwl.make_with_write_priority
    end;

  particle_list: PARTICLE_LIST;

feature

  advance_time is
    do
      write_lock;
      position.set(new_position);
      velocity.set(new_velocity);
      write_unlock;
    end;

  estimating_new_position(delta_t: DOUBLE) is
    do
      new_position.set_x(position.x + velocity.x*delta_t +
                         acceleration.x*delta_t*(delta_t/2.0));
      new_position.set_y(position.y + velocity.y*delta_t +
                         acceleration.y*delta_t*(delta_t/2.0));
    end;

  estimating_new_velocity(delta_t: DOUBLE) is
    do
      new_velocity.set_x(velocity.x + acceleration.x * delta_t);
      new_velocity.set_y(velocity.y + acceleration.y * delta_t);
    end;

  estimating_acceleration is
    local
      p: like Current;
      d,f,sin_with_x_axis,cos_with_x_axis: DOUBLE;
      i: INTEGER;
      vect: VECTOR;
    do
        -- The gravitation constant (6.67E-11) is normalized to value 1.0
      !!vect;
      fetch_external_acceleration(position.x,position.y,acceleration);
      particle_list.read_lock;
      from
        i := 1
      until
        i > particle_list.count
      loop
        p := (particle_list @ i);
        if p /= Void and then (p /= Current and not p.finish) then
          p.fetch_position(vect);
          d := position.distance(vect);
          sin_with_x_axis := (vect.y - position.y) / d;
          cos_with_x_axis := (vect.x - position.x) / d;
          f := p.fetch_mass / (d ^ 2);
          vect.set_x(f*cos_with_x_axis);
          vect.set_y(f*sin_with_x_axis);
          acceleration.add(vect);
        end;
        i := i + 1
      end;
      particle_list.read_unlock;
    end;

  estimating_advance_time_values is
    local
      delta_t: DOUBLE
    do
      read_lock;
      delta_t := time_increment * (timer.counter - last_timer_counter);
      last_timer_counter := timer.counter;

      estimating_acceleration;
      estimating_new_position(delta_t);
      estimating_new_velocity(delta_t);
      read_unlock;
    end;

feature {NONE}

  time_increment: DOUBLE; -- how much one timer click counts in time!

  is_finish: BOOLEAN;

  mass: DOUBLE;

  position: VECTOR;

  velocity: VECTOR;

  acceleration: VECTOR;

  shape: CHARACTER;

  set_shape is
    do
      if mass <= 1.0 then
        shape := '.'
      elseif mass <= 100.0 then
        shape := 'o'
      else
        shape := '0'
      end
    end;

feature

  finish: BOOLEAN is
    do
      read_lock;
      Result := is_finish;
      read_unlock
    end;

  terminate_particle is
    do
      write_lock;
      is_finish := true;
      write_unlock
    end;

  set_mass(m: DOUBLE) is
    require
      m >= 0
    do
      write_lock;
      mass := m;
      set_shape;
      write_unlock
    end;

  fetch_mass: DOUBLE is
    do
      read_lock;
      Result := mass;
      read_unlock
    end;

  fetch_position(p: VECTOR) is
      -- put particle actual position in vector 'p'
    do
      read_lock;
      p.set(position);
      read_unlock
    end;

  fetch_new_position(p: VECTOR) is
      -- put particle [possible] future position in vector 'p'
    do
      read_lock;
      p.set(new_position);
      read_unlock
    end;

  fetch_velocity(v: VECTOR) is
      -- put particle actual velocity in vector 'v'
    do
      read_lock;
      v.set(velocity);
      read_unlock
    end;

  set_velocity(v: VECTOR) is
      -- put particle actual velocity in vector 'v'
    do
      read_lock;
      velocity.set(v);
      read_unlock
    end;

feature {NONE}

  new_position: VECTOR;

  new_velocity: VECTOR;

  last_timer_counter: INTEGER;

  timer: TIMER;

  update_barrier: THREAD_BARRIER;

feature

  read_lock is
    do
      rwl.read_lock
    end;

  read_unlock is
    do
      rwl.read_unlock
    end;

  write_lock is
    do
      rwl.write_lock
    end;

  write_unlock is
    do
      rwl.write_unlock
    end;

feature {NONE}

  rwl: READ_WRITE_LOCK;

end -- PARTICLE

