--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--
-- Normalized gravitation:
--   - 1 meter = 0.5 line = 1 column
--   - the gravitation constant (6.67E-11) is normalized to 1.0
--

class MAIN

inherit
  THREAD_CONTROL;
  X11_GLOBAL

creation
  make

feature

  maximum_particles: INTEGER is 50;

  period: INTEGER is 25; -- refresh ratio in ms

  time_increment: DOUBLE is 0.3; -- simulation time increment in seconds

  num_particles: INTEGER;

  particle_list: PARTICLE_LIST;

  make is
    local
      i,update_freq: INTEGER;
      finish: BOOLEAN;
      p0,v0: VECTOR;
      inc,d,m: DOUBLE;
      part: PARTICLE;
      x11: X11_DISPLAY;
      x11_win: X11_WINDOW;
      timer: TIMER;
      barr: THREAD_BARRIER;
      clk: CLOCK;
      info: INFO;
    do
      if argument_count /= 1 then
        print_string("Usage: a.out <num-planets>%N")
      elseif not argument(1).is_integer or else argument(1).to_integer <= 0 then
        print_string("ERROR: expecting an positive integer argument!%N")
      elseif argument(1).to_integer > maximum_particles then
        print_string("ERROR: maximum number of particles exceeded (");
        print_integer(maximum_particles);
        print_string(")!%N")
      else
        num_particles := argument(1).to_integer - 1;

        !!x11.initialize;
        x11.open;
        !!x11_win.make(x11);
        x11_win.set_coordinates(10,10,910,910);
        x11_win.create_window(Void);
        x11_win.select_input(key_press_mask or exposure_mask or button_press_mask);
        x11_win.map_raised;
        x11_win.enable_flush;
        x11_win.graphic_context.set_foreground(x11_red_color(x11));

        !!particle_list.make(maximum_particles);

        !!barr.make;
        --barr.sign_on;

        !!timer.make;
        timer.set_period(period);
        timer.start;

        !!clk.make(timer,x11_win);
        clk.start;

        !!info.make(timer,x11_win);
        info.start;

        !!p0;
        !!v0;

          -- star

           -- mass
        m := 1.0E4;
           -- position
        p0.set_x(x11_win.width // 2);
        p0.set_y(x11_win.height // 2);
           -- velocity
        v0.set_zero;
        !!part.make(x11_win,m,p0,v0,time_increment,timer,barr,particle_list);
        particle_list.add(part);

          -- planets

        inc := (x11_win.height // 2).to_double / num_particles / 1.5;
        from
          d := 0.0;
          i := 1
        until
          i > num_particles
        loop
          d := d + inc;
          p0.set_x((x11_win.width // 2).to_double + d);
          p0.set_y((x11_win.height // 2).to_double);
          v0.set_x(0.0);
          v0.set_y((m / d).sqrt);
          !!part.make(x11_win,1.0,p0,v0,time_increment,timer,barr,particle_list);
          particle_list.add(part);
          i := i + 1
        end;

        particle_list.read_lock;
        from
          i := 1
        until
          i > particle_list.count
        loop
          part := particle_list @ i;
          if part /= Void then
            part.start
          end;
          i := i + 1
        end;
        particle_list.read_unlock;

        update_freq := 1000 // timer.period;
        from
        until
          finish
        loop
          timer.wait;
          x11.fetch_event;
          inspect x11.last_event
            when event_expose then
              particle_list.read_lock;
              from
                i := 1
              until
                i > particle_list.count
              loop
                part := particle_list @ i;
                if part /= Void and then not part.finish then
                  part.enable_redraw
                end;
                i := i + 1
              end;
              particle_list.read_unlock;
            when event_key_press then
              if x11.key_sym = key_sym_escape then
                finish := true
              end;
            else
          end;

          if finish then -- signal termination to all particles
            clk.terminate;
            info.terminate;
            particle_list.read_lock;
            from
              i := 1
            until
              i > particle_list.count
            loop
              part := particle_list @ i;
              if part /= Void and then not part.finish then
                part.terminate_particle
              end;
              i := i + 1
            end;
            particle_list.read_unlock;
            barr.terminate;
          else
            --barr.wait;
          end;
        end;

        join_all;
        x11_win.destroy_window;
        x11.close;
        x11.terminate
      end
    end;

end -- MAIN

