--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--
-- Changed to new global once in August 2003
--

class PARTICLE

inherit
  THREAD;
  X11_GLOBAL

creation
  make

feature {PARTICLE}

  color: X11_COLOR

  main is
    do
      update_barrier.sign_on;
      !!color.make(x11_win.display,colors @ (random_integer(colors.count-1)+1));
      x11_win.graphic_context.set_function(gx_copy);
      last_y := 0;
      last_x := 0;
      convert_to_window_coordinates;
      timer.reset_lost_periods;
      from
      until
        finish
      loop
        draw;
        timer.wait;
        estimating_advance_time_values;
        update_barrier.wait;
        advance_time;
        convert_to_window_coordinates;
      end;
      update_barrier.sign_off;
      particle_list.remove_if_present(Current)
    end;

feature -- [global] external gravity

  fetch_external_acceleration(px,py: DOUBLE;a: VECTOR) is
    do
      if px >= 0.0 and px <= x11_win.width.to_double then
        a.set_x(0.0)
      else
        if px < 0.0 then
          a.set_x((0.0 - px))
        else
          a.set_x((x11_win.width.to_double - px))
        end
      end;
      if py >= 0.0 and py/2.0 <= x11_win.height.to_double then
        a.set_y(0.0)
      else
        if py < 0.0 then
          a.set_y((0.0 - py))
        else
          a.set_y((x11_win.height.to_double - py/2.0))
        end
      end
    end;

feature -- window output

  colors: ARRAY[STRING] is
    once {"process"}
      Result := <<"red","green",
                  "blue","dark blue",
                  "brown","magenta">>;
    end;

  y,x: INTEGER;

  last_y,last_x: INTEGER;

  convert_to_window_coordinates is
      -- 1 meter = 1 y = 1 x
    do
      read_lock;
      y := position.y.rounded;
      x := position.x.rounded;
      read_unlock;
    end;

  redraw: BOOLEAN;

  enable_redraw is
    do
      redraw := true
    end;

  draw is
    do
      if redraw or y /= last_y or x /= last_x then
        x11_win.graphic_context.set_foreground(x11_win.background_color);
        x11_win.fill_circle(last_x,last_y,radius);
        x11_win.graphic_context.set_foreground(color);
        x11_win.fill_circle(x,y,radius);
        last_y := y;
        last_x := x;
        redraw := false
      end
    end;

feature

  make(x11_w: X11_WINDOW;m: DOUBLE;p0,v0: VECTOR;ti: DOUBLE;tmr: TIMER;ub: THREAD_BARRIER;pl: PARTICLE_LIST) is
    do
      x11_win := x11_w;
      mass := m;
      set_diameter;
      !!position;
      position.set(p0);
      !!velocity;
      velocity.set(v0);
      !!acceleration;
      acceleration.set_zero;
      time_increment := ti;
      timer := tmr;
      update_barrier := ub;
      particle_list := pl;
      !!new_position;
      !!new_velocity;
      is_finish := false;
      !!rwl.make_with_write_priority
    end;

  particle_list: PARTICLE_LIST;

feature

  advance_time is
    do
      write_lock;
      position.set(new_position);
      velocity.set(new_velocity);
      write_unlock;
    end;

  estimating_new_position(delta_t: DOUBLE) is
    do
      new_position.set_x(position.x + velocity.x*delta_t +
                         acceleration.x*delta_t*(delta_t/2.0));
      new_position.set_y(position.y + velocity.y*delta_t +
                         acceleration.y*delta_t*(delta_t/2.0));
    end;

  estimating_new_velocity(delta_t: DOUBLE) is
    do
      new_velocity.set_x(velocity.x + acceleration.x * delta_t);
      new_velocity.set_y(velocity.y + acceleration.y * delta_t);
    end;

  estimating_acceleration is
    local
      p: like Current;
      d,f,sin_with_x_axis,cos_with_x_axis: DOUBLE;
      i: INTEGER;
      vect: VECTOR;
    do
        -- The gravitation constant (6.67E-11) is normalized to value 1.0
      !!vect;
      fetch_external_acceleration(position.x,position.y,acceleration);
      particle_list.read_lock;
      from
        i := 1
      until
        i > particle_list.count
      loop
        p := (particle_list @ i);
        if p /= Void and then (p /= Current and not p.finish) then
          p.fetch_position(vect);
          d := position.distance(vect);
          sin_with_x_axis := (vect.y - position.y) / d;
          cos_with_x_axis := (vect.x - position.x) / d;
          f := p.fetch_mass / (d ^ 2);
          vect.set_x(f*cos_with_x_axis);
          vect.set_y(f*sin_with_x_axis);
          acceleration.add(vect);
        end;
        i := i + 1
      end;
      particle_list.read_unlock;
    end;

  estimating_advance_time_values is
    local
      delta_t: DOUBLE
    do
      read_lock;
      delta_t := time_increment * (timer.counter - last_timer_counter);
      last_timer_counter := timer.counter;

      estimating_acceleration;
      estimating_new_position(delta_t);
      estimating_new_velocity(delta_t);
      read_unlock;
    end;

feature {NONE}

  x11_win: X11_WINDOW;

  time_increment: DOUBLE; -- how much one timer click counts in time!

  is_finish: BOOLEAN;

  mass: DOUBLE;

  radius: INTEGER;

  position: VECTOR;

  velocity: VECTOR;

  acceleration: VECTOR;

  set_diameter is
    local
      d: DOUBLE
    do
      d := mass.pow(0.33333333); -- based on: vol = 4/3 PI r^3
      radius := d.rounded;
      if radius < 3 then
        radius := 3
      end
    end;

feature

  finish: BOOLEAN is
    do
      read_lock;
      Result := is_finish;
      read_unlock
    end;

  terminate_particle is
    do
      write_lock;
      is_finish := true;
      write_unlock
    end;

  set_mass(m: DOUBLE) is
    require
      m >= 0
    do
      write_lock;
      mass := m;
      set_diameter;
      write_unlock
    end;

  fetch_mass: DOUBLE is
    do
      read_lock;
      Result := mass;
      read_unlock
    end;

  fetch_position(p: VECTOR) is
      -- put particle actual position in vector 'p'
    do
      read_lock;
      p.set(position);
      read_unlock
    end;

  fetch_new_position(p: VECTOR) is
      -- put particle [possible] future position in vector 'p'
    do
      read_lock;
      p.set(new_position);
      read_unlock
    end;

  fetch_velocity(v: VECTOR) is
      -- put particle actual velocity in vector 'v'
    do
      read_lock;
      v.set(velocity);
      read_unlock
    end;

  set_velocity(v: VECTOR) is
      -- put particle actual velocity in vector 'v'
    do
      read_lock;
      velocity.set(v);
      read_unlock
    end;

feature {NONE}

  new_position: VECTOR;

  new_velocity: VECTOR;

  last_timer_counter: INTEGER;

  timer: TIMER;

  update_barrier: THREAD_BARRIER;

feature

  read_lock is
    do
      rwl.read_lock
    end;

  read_unlock is
    do
      rwl.read_unlock
    end;

  write_lock is
    do
      rwl.write_lock
    end;

  write_unlock is
    do
      rwl.write_unlock
    end;

feature {NONE}

  rwl: READ_WRITE_LOCK;

end -- PARTICLE

