-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), October 2000
--
-- Synchronization class implementing a general thread barrier.
--
-- The size of the barrier can be predefined (static) or it can
-- be defined dynamically (usign sign_on and sign_off features).
--
-- This class is protected against consecutive uses of barriers
-- (if the previous barrier is not completely finish, the attempt
-- for a new barrier will wait as long as needed, without
-- interference, and transparently to the user).
--

class THREAD_BARRIER

creation
  make,make_static

feature

  make is
    do
      number_of_threads_attr := 0;
      number_of_waiting_threads := 0;
      !!signed_threads.make;
      !!waited_threads.make;
      !!mtx.make;
      !!cvar.make;
      !!barrier_end_cvar.make;
      is_static_attr := false;
      terminated_attr := false;
    end;

  make_static(size: INTEGER) is
    require
      size > 0
    do
      number_of_threads_attr := size;
      number_of_waiting_threads := 0;
      !!waited_threads.make;
      !!mtx.make;
      !!cvar.make;
      !!barrier_end_cvar.make;
      is_static_attr := true;
      terminated_attr := false;
    end;

  terminated: BOOLEAN is
    do
      mtx.lock;
      Result := terminated_attr;
      mtx.unlock;
    end;

  terminate is
    do
      mtx.lock;
      number_of_threads_attr := 0;
      number_of_waiting_threads := 0;
      terminated_attr := true;
      cvar.broadcast;
      mtx.unlock;
    end;

  release is
      -- all waiting threads in barrier will be released.
    do
      mtx.lock;
      cvar.broadcast;
      mtx.unlock;
    end;

  is_static: BOOLEAN is
    -- is the size of the barrier fixed?
    do
      mtx.lock;
      Result := is_static_attr;
      mtx.unlock;
    end;

  set_number_of_threads(size: INTEGER) is
    require
      is_static;
      size > 0;
    do
      mtx.lock;
      check number_of_waiting_threads = 0 end;
      number_of_threads_attr := size;
      mtx.unlock
    end;

  number_of_threads: INTEGER is
    -- number of signed threads
    do
      mtx.lock;
      Result := number_of_threads_attr;
      mtx.unlock;
    end;

  signed: BOOLEAN is
      -- is calling thread already signed?
    require
      not is_static
    do
      mtx.lock;
      Result := (thread_id_index(tid,signed_threads) > 0);
      mtx.unlock
    end;

  sign_on is
      -- calling thread will be a new user of barrier
    require
      not is_static;
      not signed
    do
      mtx.lock;
      signed_threads.add_last(tid);
      number_of_threads_attr := number_of_threads_attr + 1;
      mtx.unlock
    end;

  sign_off is
      -- calling thread won't be a user of barrier anymore
    require
      not is_static;
      signed
    do
      mtx.lock;
      signed_threads.remove(thread_id_index(tid,signed_threads));
      number_of_threads_attr := number_of_threads_attr - 1;
      mtx.unlock
    end;

  wait is
      -- Calling thread will wait until 'number_of_threads' threads
      --  are waiting (then they will all unblock).
      -- On termination initializes new barrier (with the same threads
      --  if the barrier is dynamic)
    require
      is_static or else signed
    do
      mtx.lock;
      if not terminated_attr then
        if (thread_id_index(tid,waited_threads) > 0) then
          barrier_end_cvar.wait(mtx) -- wait until last barrier terminates
        end;
        if not terminated_attr then
          number_of_waiting_threads := number_of_waiting_threads + 1;
          if number_of_waiting_threads = number_of_threads_attr then
            number_of_waiting_threads := number_of_waiting_threads - 1;
            if number_of_waiting_threads > 0 then
              waited_threads.add_last(tid);
              cvar.signal
            end
          else
            cvar.wait(mtx);
            number_of_waiting_threads := number_of_waiting_threads - 1;
              if number_of_waiting_threads > 0 then
              waited_threads.add_last(tid);
              cvar.signal
            else
              waited_threads.clear;
              barrier_end_cvar.broadcast -- notify end of barrier
            end;
          end;
        end;
      end;
      mtx.unlock;
    end;

feature {THREAD_BARRIER}

  thread_id_index(thr_id: THREAD_ID;list: LINKED_LIST[THREAD_ID]): INTEGER is
      --  returns 0 if not found
    local
      i: INTEGER;
      finish: BOOLEAN
    do
      from
        i := 1
      until
        finish or else i > list.count
      loop
        if (list @ i).same_as(thr_id) then
          Result := i;
          finish := true
        end;
        i := i + 1
      end;
    end;

  tid: THREAD_ID is
      -- one per-thread as necessary
    once
      !!Result.make
    end;

feature {NONE}

  terminated_attr: BOOLEAN;

  is_static_attr: BOOLEAN;

  number_of_threads_attr: INTEGER;

  number_of_waiting_threads: INTEGER;

  signed_threads: LINKED_LIST[THREAD_ID];

  waited_threads: LINKED_LIST[THREAD_ID];

  mtx: MUTEX;

  cvar: CONDITION_VARIABLE;

  barrier_end_cvar: CONDITION_VARIABLE;

end -- THREAD_BARRIER

