-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), October 2000
--

class THREAD_CONTROL

feature

  running: BOOLEAN is
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      Result := pthread_running(pthread_data)
    end;

  detached: BOOLEAN is
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      Result := pthread_detached(pthread_data)
    end;

  is_same_thread: BOOLEAN is
      -- is the calling thread the same as the owner of current object?
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      if running then
        Result := c_is_same_thread(pthread_data)
      end
    end;

  is_main_thread,is_root_thread: BOOLEAN is
      -- are we in main (root) thread?
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      Result := c_is_root_thread(pthread_data)
    end;

  detach is
      -- detach the owner thread of current object
    require
      running and then is_same_thread;
      not detached
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      c_detach_pthread(pthread_data);
    end;

  exit is
      -- forces termination of the owner thread of current object
    require
      running and then is_same_thread
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      c_exit_pthread(pthread_data);
    end;

  join is
      -- The caller will block while the owner thread of current object
      --  is running
    require
      not detached;
      not running or else not is_same_thread
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      c_join_pthread(pthread_data)
    end;

  join_all_childs,join_all is
      -- The caller will block while all direct child threads of the
      --  owner of current object are running
      -- Ignores detached direct childs.
      -- This feature is usable by the thread owning Current object
      --  (unlike 'join' feature).
      -- Returns immediately if there isn't any child.
    do
      if pthread_data.is_null then
        fetch_thread_data
      end;
      c_join_all_childs_pthread(pthread_data)
    end;

feature {THREAD_CONTROL}

  pthread_data: POINTER;
    -- internal pthread data

feature {NONE}

  fetch_thread_data is
    do
      pthread_data := c_get_current_thread_data;
    end;

  frozen c_get_current_thread_data: POINTER is
    external "C"
    alias "get_current_thread_data"
    end;

  frozen pthread_running(td: POINTER): BOOLEAN is
    external "C"
    end;

  frozen pthread_detached(td: POINTER): BOOLEAN is
    external "C"
    end;

  frozen pthread_error(td: POINTER): BOOLEAN is
    external "C"
    end;

  frozen c_is_same_thread(td: POINTER): BOOLEAN is
    external "C"
    alias "is_same_thread"
    end;

  frozen c_is_root_thread(td: POINTER): BOOLEAN is
    external "C"
    alias "is_root_thread"
    end;

  frozen c_detach_pthread(td: POINTER) is
    external "C"
    alias "detach_pthread"
    end;

  frozen c_exit_pthread(td: POINTER) is
    external "C"
    alias "exit_pthread"
    end;

  frozen c_join_pthread(td: POINTER) is
    external "C"
    alias "join_pthread"
    end;

  frozen c_join_all_childs_pthread(td: POINTER) is
    external "C"
    alias "join_all_childs_pthread"
    end;

invariant

  not pthread_error(pthread_data)

end -- THREAD_CONTROL

