--          This file is part of SmallEiffel The GNU Eiffel Compiler.
--       Copyright (C) 1994-2002 LORIA - INRIA - U.H.P. Nancy 1 - FRANCE
--          Dominique COLNET and Suzanne COLLIN - SmallEiffel@loria.fr
--                       http://SmallEiffel.loria.fr
-- SmallEiffel is  free  software;  you can  redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free
-- Software  Foundation;  either  version  2, or (at your option)  any  later
-- version. SmallEiffel is distributed in the hope that it will be useful,but
-- WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License
-- for  more  details.  You  should  have  received a copy of the GNU General
-- Public  License  along  with  SmallEiffel;  see the file COPYING.  If not,
-- write to the  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-- Boston, MA 02111-1307, USA.
--
-- Changed for thread-safety, by Miguel Oliveira e Silva (mos@det.ua.pt)
--
class ONCE_ROUTINE_POOL
   --
   -- Singleton object in charge of runnable once routines.
   -- This singleton is shared via the GLOBALS.`once_routine_pool' once function.
   --

inherit GLOBALS;

feature {RUN_FEATURE} -- For `compile_to_c' as well as `compile_to_jvm' :

   register_procedure(rf: RUN_FEATURE) is
      require
         rf.is_once_procedure
      do
         check
            not procedure_list.has(rf)
         end;
         procedure_list.add_last(rf);
      end;

   register_function(rf: RUN_FEATURE) is
      require
         rf.is_once_function
      do
         check
            not function_list.has(rf)
         end;
         function_list.add_last(rf);
      end;

feature {GC_HANDLER}

   gc_mark_in(c_code: STRING) is
	 -- To mark results of once functions because they are part
	 -- of the root.
      local
         i: INTEGER;
         rf: RUN_FEATURE;
         of: ONCE_FUNCTION;
         mem: FIXED_ARRAY[ONCE_FUNCTION];
         rt: E_TYPE;
         oresult: STRING;
      do
         if function_list.count > 0 then
            from
               i := function_list.upper;
               !!mem.with_capacity(i // 2);
            until
               i < 0
            loop
               rf := function_list.item(i);
               of ?= rf.base_feature;
	       check
		  of /= Void
	       end;
               if not mem.fast_has(of) then
                  if rf.current_type.run_class.at_run_time then
                     mem.add_last(of);
                     rt := rf.result_type;
                     if rt.need_gc_mark_function then
                        oresult := o_result(of);
			gc_handler.mark_for(c_code,oresult,rt.run_class);
                     end;
                  end;
               end;
               i := i - 1;
            end;
         end;
      end;

feature {JVM}

   fields_count: INTEGER;

   jvm_define_fields is
      local
         byte_idx, idx_flag, i: INTEGER;
         rf: RUN_FEATURE;
         bf: E_FEATURE;
         name_list: FIXED_ARRAY[INTEGER];
      do
         !!name_list.with_capacity(fields_count);
         if function_list.count > 0 then
            from
               i := function_list.upper;
               byte_idx := constant_pool.idx_utf8(fz_41);
            until
               i < 0
            loop
               rf := function_list.item(i);
               bf := rf.base_feature;
               idx_flag := idx_name_for_flag(bf);
               if name_list.fast_has(idx_flag) then
               else
                  name_list.add_last(idx_flag);
                  -- ---------- Static field for flag :
                  field_info.add(9,idx_flag,byte_idx);
                  -- ---------- Static field for result :
                  field_info.add(9,
                                 idx_name_for_result(bf),
                                 idx_descriptor(rf.result_type.run_type));
               end;
               i := i - 1;
            end;
         end;
         if procedure_list.count > 0 then
            from
               i := procedure_list.upper;
               byte_idx := constant_pool.idx_utf8(fz_41);
            until
               i < 0
            loop
               rf := procedure_list.item(i);
               bf := rf.base_feature;
               idx_flag := idx_name_for_flag(bf);
               if name_list.fast_has(idx_flag) then
               else
                  name_list.add_last(idx_flag);
                  -- ---------- Static field for flag :
                  field_info.add(9,idx_flag,byte_idx);
               end;
               i := i - 1;
            end;
         end;
      end;

   jvm_initialize_fields is
      local
         i: INTEGER;
      do
         if jvm_flag_list.count > 0 then
            from
               i := jvm_flag_list.upper;
            until
               i < 0
            loop
               -- Set once flag :
               code_attribute.opcode_iconst_0;
               code_attribute.opcode_putstatic(jvm_flag_list.item(i),-1);
               i := i - 1;
            end;
         end;
      end;

feature {RUN_FEATURE} -- For `compile_to_jvm' only :

   idx_fieldref_for_flag(rf: RUN_FEATURE): INTEGER is
      require
         rf.is_once_routine
      local
	 cp: CONSTANT_POOL;
      do
	 cp := constant_pool;
         Result := cp.idx_fieldref3(jvm_root_class,
				    o_flag(rf.base_feature),
				    fz_41);
      end;

feature {NONE} -- For `compile_to_c' as well as `compile_to_jvm' :

   idx_fieldref_for_result(rf: RUN_FEATURE): INTEGER is
      require
         rf.is_once_function
      local
	 cp: CONSTANT_POOL;
	 rt: E_TYPE;
      do
	 cp := constant_pool;
         rt := rf.result_type.run_type;
         Result := cp.idx_fieldref3(jvm_root_class,
				    o_result(rf.base_feature),
				    jvm_descriptor(rt));
      end;

   procedure_list: FIXED_ARRAY[RUN_FEATURE] is
	 -- Live set of once procedures.
      once
         !!Result.with_capacity(32);
      end;

   function_list: FIXED_ARRAY[RUN_FEATURE] is
	 -- Live set of once functions.
      once
         !!Result.with_capacity(32);
      end;

   o_flag(bf: E_FEATURE): STRING is
	 -- Compute the only one corresponding `flag': fBCxxKey used
	 -- to check that execution is done.
      require
	 bf /= Void
      do
	 Result := o_flag_buffer;
         Result.clear;
         Result.append(once "fBC");
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   buffer: STRING is
      once
	 !!Result.make(128);
      end;

   body: STRING is 
      once
	 !!Result.make(128);
      end;

   c_define_o_result_buffer: STRING is
      once
	 !!Result.make(64);
      end;

   o_flag_buffer: STRING is
      once
	 !!Result.make(32);
      end;

   o_result_buffer: STRING is
      once
	 !!Result.make(32);
      end;

   jvm_descriptor_buffer: STRING is
      once
	 !!Result.make(32);
      end;

feature {ABSTRACT_RESULT,RUN_FEATURE}

   o_result(bf: E_FEATURE): STRING is
      require
	 bf /= Void
      do
	 Result := o_result_buffer;
         Result.clear;
         Result.append(once "oBC");
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   c_put_o_result(rf: RUN_FEATURE) is
      require
	 rf.is_once_function
      do
	 cpp.put_string(o_result(rf.base_feature));
      end;

   jvm_result_load(rf: RUN_FEATURE) is
      require
	 rf.is_once_function
      local
         result_space, idx_result: INTEGER;
      do
         result_space := rf.result_type.jvm_stack_space;
         idx_result := idx_fieldref_for_result(rf);
         code_attribute.opcode_getstatic(idx_result,result_space);
      end;

   jvm_result_store(rf: RUN_FEATURE) is
      require
	 rf.is_once_function
      local
         result_space, idx_result: INTEGER;
      do
         result_space := rf.result_type.jvm_stack_space;
         idx_result := idx_fieldref_for_result(rf);
         code_attribute.opcode_putstatic(idx_result,- result_space);
      end;

feature {C_PRETTY_PRINTER} -- For `compile_to_c' :

   c_pre_compute is
	 -- Generate C code to pre_compute some functions.
      local
	 i: INTEGER;
	 memory: FIXED_ARRAY[E_FEATURE];
	 rf: RUN_FEATURE;
	 bf: E_FEATURE;
      do
	 echo.put_string(fz_04);
	 echo.put_string(fz_05);
	 cpp.put_string(fz_pco_comment);
	 from
	    i := function_list.upper;
	    !!memory.with_capacity(64);
	 until
	    i < 0
	 loop
	    rf := function_list.item(i);
	    if rf.run_class.at_run_time then
	       if rf.is_pre_computable then
		  bf := rf.base_feature;
		  if not memory.fast_has(bf) then
		     memory.add_last(bf);
		     echo.put_character('%T');
		     echo.put_string(bf.base_class.name.to_string);
		     echo.put_character('.');
		     echo.put_string(rf.name.to_string);
		     echo.put_character('%N');
		     c_pre_compute_of(rf,bf);
		     cpp.put_string(fz_pco_comment);
		  end;
	       end;
	    end;
	    i := i - 1;
	 end;
	 echo.print_count(fz_04,memory.count);
      end;

feature {RUN_FEATURE} -- For `compile_to_c' :

   c_define_o_flag(rf: RUN_FEATURE) is
	 -- Add the definition/initialization of the corresponding
	 -- `o_flag' if not yet done.
      require
	 rf.is_once_routine
      local
	 bf: E_FEATURE;
         ort: ONCE_ROUTINE;
	 bcbf: BASE_CLASS;
	 flag: STRING;
	 index: STRING;
	 mutex: STRING;
	 typedef: STRING;
         i: INTEGER;
	 rt: E_TYPE;
      do
	 bf := rf.base_feature;
	 bcbf := bf.base_class;
	 flag := o_flag(bf);
	 if not bcbf.once_flag(flag) then
            if ace.pthread then
                  -- Insertion of all the necessary code outside once routines

               declare_once_keys;
	       index := o_index(bf);
	       mutex := o_mutex(bf);
	       typedef := o_typedef(bf);
               ort ?= bf;
               check ort /= Void end;

               if ort.thread_scope then
                  if ort.class_thread_scope then
                     buffer.copy(once "%N#define ");
                     buffer.append(index);
                     buffer.extend(' ');
                     buffer.append(num_of_thread_once_routines.to_string);
                     buffer.extend('%N');
                     cpp.put_string_on_h(buffer);
                     num_of_thread_once_routines := num_of_thread_once_routines + 1;
                  end;
                  if rf.is_once_function or ort.key_scope then
                     buffer.copy(once "typedef struct%N{%N");
                     if rf.is_once_function then
                        rt := rf.result_type;
                        buffer.append(once "  T");
                        if rt.is_expanded then
                           rt.id.append_in(buffer);
                           buffer.extend(' ');
                        else
                           buffer.extend('0');
                           buffer.extend('*');
                        end;
                        buffer.append(once "result;%N");
                        cpp.put_string_on_h(buffer);
                        buffer.clear;
                     end;
                     if ort.key_scope then
                        from
                           i := ort.once_key_list.lower
                        until
                           i > ort.once_key_list.upper
                        loop
                           buffer.append(once "  int ");
                           buffer.append(o_name_from_key(bf,ort.once_key_list @ i));
                           buffer.append(";%N");
                           cpp.put_string_on_h(buffer);
                           buffer.clear;
                           i := i + 1
                        end
                     end;
                     buffer.append(once "} ");
                     buffer.append(typedef);
                     buffer.append(once ";%N");
                     cpp.put_string_on_h(buffer);
                  end
               else -- ort.process_scope
                  if ort.class_process_scope then
	             buffer.copy(once "int ");
	             buffer.append(flag);
	             cpp.put_extern2(buffer,'0');
	             buffer.copy(once "pthread_mutex_t ");
	             buffer.append(mutex);
	             cpp.put_extern5(buffer,once "PTHREAD_MUTEX_INITIALIZER");
                     if ort.key_scope then
                        from
                           i := ort.once_key_list.lower
                        until
                           i > ort.once_key_list.upper
                        loop
                           cpp.put_string(once "int ");
                           cpp.put_string(o_name_from_key(bf,ort.once_key_list @ i));
                           cpp.put_string(" = 0;%N");
                           i := i + 1
                        end
                     end
                  end
               end
            else
	       buffer.copy(once "int ");
	       buffer.append(flag);
	       cpp.put_extern2(buffer,'0');
            end
	 end;
      end;

   c_define_o_result(rf: RUN_FEATURE) is
      require
	 rf.is_once_function
      local
	 bf: E_FEATURE;
         ort: ONCE_ROUTINE;
	 bcbf: BASE_CLASS;
	 oresult: STRING;
	 rt: E_TYPE;
      do
	 bf := rf.base_feature;
         ort ?= bf;
         check ort /= Void end;
            -- only the result of class process once functions is handled here.
         if not ace.pthread or else ort.class_process_scope then
	    bcbf := bf.base_class;
	    oresult := o_result(bf);
	    if not bcbf.once_flag(oresult) then
	       buffer.clear;
	       buffer.extend('T');
	       rt := rf.result_type;
	       if rt.is_expanded then
	          rt.id.append_in(buffer);
	          buffer.extend(' ');
	       else
	          buffer.extend('0');
	          buffer.extend('*');
	       end;
	       buffer.append(oresult);
	       c_define_o_result_buffer.clear;
	       rt.c_initialize_in(c_define_o_result_buffer);
	       cpp.put_extern5(buffer,c_define_o_result_buffer);
	    end;
	 end
     end;

   c_test_o_flag(rf: RUN_FEATURE) is
      require
	 rf.is_once_routine
      local
	 flag: STRING;
	 bf: E_FEATURE;
         ort: ONCE_ROUTINE;
	 index: STRING;
	 typedef: STRING;
	 mutex: STRING;
	 external_result: STRING;
	 oresult: STRING;
         i: INTEGER;
	 rt: E_TYPE;
      do
         bf := rf.base_feature;
	 flag := o_flag(rf.base_feature);
         if ace.pthread then
            index := o_index(bf);
            typedef := o_typedef(bf);
            mutex := o_mutex(bf);
            external_result := o_external_result(bf);
            oresult := o_result(bf);
            ort ?= bf;
            check ort /= Void end;

            cpp.put_string(once "int _onceness_status_;%N");
            if ort.thread_scope then
               if rf.is_once_function or ort.key_scope then
                  cpp.put_string(typedef);
                  cpp.put_string(once " *");
                  cpp.put_string(external_result);
                  cpp.put_string(once ";%N");
               end;
               if rf.is_once_function then
                  buffer.clear;
                  buffer.extend('T');
                  rt := rf.result_type;
                  if rt.is_expanded then
	             rt.id.append_in(buffer);
	             buffer.extend(' ');
	          else
	             buffer.extend('0');
	             buffer.extend('*');
	          end;
	          buffer.append(oresult);
	          c_define_o_result_buffer.clear;
	          rt.c_initialize_in(c_define_o_result_buffer);

	          cpp.put_string(buffer);
	          cpp.put_string(once "=");
                  cpp.put_string(c_define_o_result_buffer);
	          cpp.put_string(once ";%N");
               end;

               if ort.object_thread_scope then
                  cpp.put_string(once "_set_index_object_thread_once_routine(&C->");
                  cpp.put_string(index);
                  cpp.put_string(once ");%N");
               end;

               if rf.is_once_function or ort.key_scope then
                  cpp.put_string(external_result);
	          cpp.put_string(once " = (");
                  cpp.put_string(typedef);
	          cpp.put_string(once "*)_get_result_of_thread_once_routine_(");
                  if ort.class_thread_scope then
                    cpp.put_string(once "0,");
                  else
                    cpp.put_string(once "1,C->");
                  end;
                  cpp.put_string(index);
	          cpp.put_string(once ");%N");
                  cpp.put_string(once "if(");
                  cpp.put_string(external_result);
                  cpp.put_string(once " == NULL){%N");
                  cpp.put_string(once "_onceness_status_ = 0;%N");
                  cpp.put_string(external_result);
                  cpp.put_string(once " = se_malloc(sizeof(");
                  cpp.put_string(typedef);
                  cpp.put_string(once "));%N");
	          cpp.put_string(once "_set_result_of_thread_once_routine_(");
                  if ort.class_thread_scope then
                    cpp.put_string(once "0,");
                  else
                    cpp.put_string(once "1,C->");
                  end;
                  cpp.put_string(index);
	          cpp.put_string(once ",(void*)");
                  cpp.put_string(external_result);
                  cpp.put_string(once ");%N");
                  cpp.put_string(once "}else%N");
                  cpp.put_string(once "_onceness_status_ = 1;%N");
               else
	          cpp.put_string(once "if(_is_first_thread_once_routine_call_(");
                  if ort.class_thread_scope then
                    cpp.put_string(once "0,");
                  else
                    cpp.put_string(once "1,C->");
                  end;
                  cpp.put_string(index);
	          cpp.put_string(once "))");
                  cpp.put_string(once "_onceness_status_ = 0;%N");
                  cpp.put_string(once "else%N");
                  cpp.put_string(once "_onceness_status_ = 1;%N");
               end;

               cpp.put_string(once "if(_onceness_status_ == 0");
               if ort.key_scope then
                  cpp.put_string(once "|| (");
                  cpp.put_string(once "(");
                  from
                     i := ort.once_key_list.lower
                  until
                     i > ort.once_key_list.upper
                  loop
                     if i > ort.once_key_list.lower then
                        cpp.put_string(once "||%N   ");
                     end;
                     cpp.put_string(once "_is_first_once_key_routine_call_(%"");
                     cpp.put_string(ort.once_key_list @ i);
                     cpp.put_string(once "%",");
                     cpp.put_string(external_result);
                     cpp.put_string(once "->");
                     cpp.put_string(o_name_from_key(bf,ort.once_key_list @ i));
                     cpp.put_string(once ")");
                     i := i + 1
                  end;
                  cpp.put_string(once "))")
               end;
               cpp.put_string(once "){");
               cpp.put_string(once "_set_first_thread_once_routine_call_(");
               if ort.class_thread_scope then
                  cpp.put_string(once "0,");
               else
                  cpp.put_string(once "1,C->");
               end;
               cpp.put_string(index);
               cpp.put_string(once ");%N");
               if ort.key_scope then
                  from
                     i := ort.once_key_list.lower
                  until
                     i > ort.once_key_list.upper
                  loop
                     cpp.put_string(once "_set_once_key_routine_called_(%"");
                     cpp.put_string(ort.once_key_list @ i);
                     cpp.put_string(once "%",&");
                     cpp.put_string(external_result);
                     cpp.put_string(once "->");
                     cpp.put_string(o_name_from_key(bf,ort.once_key_list @ i));
                     cpp.put_string(once ");%N");
                     i := i + 1
                  end
               end;
               cpp.put_string(once "{");
               if rf.is_once_function then
                  cpp.put_string(once "%N");
               else
                  cpp.put_string(once "{%N");
               end
            else -- ort.process_scope
               if rf.is_once_function and ort.object_process_scope then
                  buffer.clear;
                  buffer.extend('T');
                  rt := rf.result_type;
                  if rt.is_expanded then
	             rt.id.append_in(buffer);
	             buffer.extend(' ');
	          else
	             buffer.extend('0');
	             buffer.extend('*');
	          end;
	          buffer.append(oresult);
	          c_define_o_result_buffer.clear;
	          rt.c_initialize_in(c_define_o_result_buffer);
	          cpp.put_string(buffer);
	          cpp.put_string(once "=");
                  cpp.put_string(c_define_o_result_buffer);
	          cpp.put_string(once ";%N");
               end;
               cpp.put_string(once "pthread_mutex_lock(&");
               if ort.object_process_scope then
                  cpp.put_string(once "C->")
               end;
               cpp.put_string(mutex);
               cpp.put_string(once ");%N");
               cpp.put_string(once "_onceness_status_ = ");
               if ort.object_process_scope then
                  cpp.put_string(once "C->")
               end;
               cpp.put_string(flag);
               cpp.put_string(once ";%N");
               if ort.object_process_scope then
                  cpp.put_string(once "C->")
               end;
               cpp.put_string(flag);
	       cpp.put_string(once " = 1;%N");
               cpp.put_string(once "pthread_mutex_unlock(&");
               if ort.object_process_scope then
                  cpp.put_string(once "C->")
               end;
               cpp.put_string(mutex);
               cpp.put_string(once ");%N");

               cpp.put_string(once "if(_onceness_status_ == 0");
               if ort.key_scope then
                  cpp.put_string(once " ||%N   ");
                  from
                     i := ort.once_key_list.lower
                  until
                     i > ort.once_key_list.upper
                  loop
                     if i > ort.once_key_list.lower then
                        cpp.put_string(once " ||%N   ");
                     end;
                     cpp.put_string(once "_is_first_once_key_routine_call_(%"");
                     cpp.put_string(ort.once_key_list @ i);
                     cpp.put_string(once "%",");
                     if ort.object_process_scope then
                        cpp.put_string(once "C->")
                     end;
                     cpp.put_string(o_name_from_key(bf,ort.once_key_list @ i));
                     i := i + 1
                  end;
                  cpp.put_string(once ")){")
                  from
                     i := ort.once_key_list.lower
                  until
                     i > ort.once_key_list.upper
                  loop
                     cpp.put_string(once "_set_once_key_routine_called_(%"");
                     cpp.put_string(ort.once_key_list @ i);
                     cpp.put_string(once "%",&");
                     if ort.object_process_scope then
                        cpp.put_string(once "C->")
                     end;
                     cpp.put_string(o_name_from_key(bf,ort.once_key_list @ i));
                     cpp.put_string(once ");%N");
                     i := i + 1
                  end
                  cpp.put_string(once "{%N")
               else
                  cpp.put_string(once "){{%N")
               end;
            end
         else
            cpp.put_string(once "if(");
            cpp.put_string(flag);
            cpp.put_string(once "==0){");
            cpp.put_string(flag);
	    cpp.put_string(once "=1;{%N");
         end
      end;

   c_return_o_result(rf: RUN_FEATURE) is
      require
	 rf.is_once_function
      local
	 bf: E_FEATURE;
         ort: ONCE_ROUTINE;
	 flag: STRING;
	 oresult: STRING;
         external_result: STRING;
         typedef: STRING;
	 mutex: STRING;
         index: STRING;
      do
	 bf := rf.base_feature;
         if ace.pthread then
               -- set/fetch once function result

	    flag := o_flag(rf.base_feature);
            oresult := o_result(bf);
            external_result := o_external_result(bf);
            typedef := o_typedef(bf);
            mutex := o_mutex(bf);
            index := o_index(bf);
            ort ?= bf;
            check ort /= Void end;

            if ort.thread_scope then
               cpp.put_string(external_result);
               cpp.put_string(once "->result = ");
               cpp.put_string(oresult);
               cpp.put_string(once ";%N}}");
               cpp.put_string(once "%Nelse%N{%N");
               cpp.put_string(oresult);
               cpp.put_string(once " = ");
               cpp.put_string(external_result);
               cpp.put_string(once "->result;%N");
               cpp.put_string(once "}%Nreturn ");
            elseif ort.object_process_scope then
               cpp.put_string(once "pthread_mutex_lock(&C->");
               cpp.put_string(mutex);
               cpp.put_string(once ");%N");
               cpp.put_string(once "C->");
               cpp.put_string(oresult);
               cpp.put_string(once " = ");
               cpp.put_string(oresult);
               cpp.put_string(once ";%N");
               cpp.put_string(once "pthread_mutex_unlock(&C->");
               cpp.put_string(mutex);
               cpp.put_string(once ");%N");
               cpp.put_string(once "}}%Nelse%N{%N");
               cpp.put_string(once "pthread_mutex_lock(&C->");
               cpp.put_string(mutex);
               cpp.put_string(once ");%N");
               cpp.put_string(oresult);
               cpp.put_string(once " = ");
               cpp.put_string(once "C->");
               cpp.put_string(oresult);
               cpp.put_string(once ";%N");
               cpp.put_string(once "pthread_mutex_unlock(&C->");
               cpp.put_string(mutex);
               cpp.put_string(once ");%N");
               cpp.put_string(once "}%Nreturn ");
            else -- ort.class_process_scope
               cpp.put_string(once "}}%Nreturn ");
            end
         else
            cpp.put_string(once "}}%Nreturn ");
         end
         cpp.put_string(o_result(rf.base_feature));
         cpp.put_string(once ";%N}%N");
      end;

feature {NONE} -- For `compile_to_jvm' :

   jvm_flag_list: FIXED_ARRAY[INTEGER] is
      once
         !!Result.with_capacity(32);
      end;

   idx_descriptor(rt: E_TYPE): INTEGER is
      require
         rt /= Void
      do
         Result := constant_pool.idx_utf8(jvm_descriptor(rt));
      end

   idx_name_for_result(bf: E_FEATURE): INTEGER is
      require
         bf /= Void
      do
         Result := constant_pool.idx_utf8(o_result(bf));
      end

   idx_name_for_flag(bf: E_FEATURE): INTEGER is
      require
         bf /= Void
      do
         Result := constant_pool.idx_utf8(o_flag(bf));
      end

   jvm_descriptor(rt: E_TYPE): STRING is
      do
	 Result := jvm_descriptor_buffer;
         Result.clear;
         if rt.is_reference then
            Result.append(jvm_root_descriptor)
         else
            rt.jvm_descriptor_in(Result)
         end;
      end;

feature {RUN_FEATURE_6,RUN_FEATURE_11}

   is_pre_computable(rf: RUN_FEATURE): BOOLEAN is
      require
	 rf.is_once_function
      do
            -- attempt to make all once routines alike (to guarantee
            -- a correct code generation).
         if not ace.pthread then
            if frozen_general.fast_has(rf.name.to_string) then
               Result := true;
            elseif rf.arguments = Void and then not rf.use_current then
               if rf.routine_body = Void then
                  Result := true;
               elseif not ace.invariant_check then
                  Result := rf.routine_body.is_pre_computable;
               end;
            end;
         end;
         if Result then
            if rf.require_assertion /= Void then
               rf.require_assertion.clear_run_feature;
            end;
            if rf.ensure_assertion /= Void then
               rf.ensure_assertion.clear_run_feature;
            end;
         end;
      end;

feature {NONE}

   c_pre_compute_of(rf: RUN_FEATURE; bf: E_FEATURE) is
      require
	 rf.is_once_function;
         rf.is_pre_computable;
         cpp.on_c;
      local
	 result_type: E_TYPE; local_vars: LOCAL_VAR_LIST; rf3: RUN_FEATURE_3;
      do
	 if rf.require_assertion /= Void then
	    rf.require_assertion.compile_to_c;
	 end;
	 result_type := rf.result_type;
	 if result_type.is_user_expanded then
	    rf3 := result_type.run_class.a_default_create;
	    if rf3 /= Void then
	       cpp.put_proc_call_0(rf3,Void,o_result(rf.base_feature));
	    end;
	 end;
	 --
	 local_vars := rf.local_vars;
         if local_vars /= Void then
            cpp.put_character('{');
            local_vars.c_declare;
         end;
	 --
         if rf.routine_body /= Void then
            rf.routine_body.compile_to_c;
         end;
	 --
	 if rf.ensure_assertion /= Void then
	    rf.ensure_assertion.compile_to_c;
         end;
	 --
         if local_vars /= Void then
            cpp.put_character('}');
         end;
      end;

   frozen_general: ARRAY[STRING] is
      once
         if ace.pthread then
               -- attempt to make all once routines alike (to guarantee
               -- a correct code generation).
            Result := <<"">>;
         else
            Result := <<as_std_error, as_std_input, as_io, as_std_output>>;
         end
      end;

   fz_pco_comment: STRING is "/*PCO*/%N";

   singleton_memory: ONCE_ROUTINE_POOL is
      once
	 Result := Current;
      end;

feature {TMP_FEATURE} -- pthread related attributes:

   register_once_key(k: STRING) is
      do
        if not once_key_list.has(k) then
           once_key_list.add_last(k)
        end
      end;

feature {NONE} -- pthread related features:

   declare_once_keys is
      local
         i: INTEGER
      once
	 cpp.put_extern5(once "pthread_mutex_t _once_key_list_mutex_",
                         once "PTHREAD_MUTEX_INITIALIZER");
         cpp.put_string_on_h(once "%Ntypedef struct {char *key;int count;} _once_key_type_;%N");
         cpp.put_string_on_h(once "extern _once_key_type_ _once_key_list_[];%N");
         cpp.put_string_on_h(once "#define _TSSE_INCLUDE_ONCE_KEY_ROUTINES_%N");
         cpp.put_string(once "_once_key_type_ _once_key_list_[] = {");
         from
            i := once_key_list.lower
         until
            i > once_key_list.upper
         loop
            if i > once_key_list.lower then
               cpp.put_string(once ",")
            end;
            cpp.put_string(once "{%"");
            cpp.put_string(once_key_list @ i);
            cpp.put_string(once "%",0}");
            i := i + 1
         end;
         cpp.put_string(once "};%N")
	 buffer.clear;
         once_key_list.count.append_in(buffer);
	 cpp.put_extern5(once "int _once_key_list_size_",buffer);
      end;

   once_key_list: FIXED_ARRAY[STRING] is
      once
         !!Result.with_capacity(8);
      end;

   num_of_thread_once_routines: INTEGER;

   o_mutex(bf: E_FEATURE): STRING is
	 -- Compute the only one corresponding `mutex': mBCxxKey
      require
	 bf /= Void
      do
	 Result := o_pthread_buffer;
         Result.clear;
         Result.append(once "mBC");
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   o_index(bf: E_FEATURE): STRING is
	 -- Compute the only one corresponding `index': iBCxxKey
      require
	 bf /= Void
      do
	 Result := o_pthread_buffer;
         Result.clear;
         Result.append(once "iBC");
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   o_typedef(bf: E_FEATURE): STRING is
	 -- Compute the only one corresponding result `typedef': tBCxxKey
      require
	 bf /= Void
      do
	 Result := o_pthread_buffer;
         Result.clear;
         Result.append(once "tBC");
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   o_external_result(bf: E_FEATURE): STRING is
	 -- Compute the only one corresponding external result : eBCxxKey
      require
	 bf /= Void
      do
	 Result := o_pthread_buffer;
         Result.clear;
         Result.append(once "eBC");
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   o_name_from_key(bf: E_FEATURE;key: STRING): STRING is
      require
	 key /= Void
      local
         i: INTEGER;
         c: CHARACTER
      do
	 Result := o_pthread_buffer;
         Result.copy(key);
         from
            i := key.lower
         until
            i > key.upper
         loop
            c := key @ i;
            if not c.is_letter and not c.is_digit then
               Result.put('_',i)
            end;
            i := i + 1
         end;
         bf.base_class.id.append_in(Result);
         Result.append(bf.first_name.to_key);
         Result.append(once "_key_count");
	 Result := string_aliaser.item(Result);
      ensure
	 Result = string_aliaser.item(Result)
      end;

   o_pthread_buffer: STRING is
      once
	 !!Result.make(32);
      end;

invariant

   is_real_singleton: Current = singleton_memory

end -- ONCE_ROUTINE_POOL

