/* c_thread.c
** This  file  is   free  software,  which   comes  along  with   thread-safe
** SmallEiffel. This software  is distributed  in the  hope that  it will  be
** useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
** you want, provided this  header is kept unaltered,  and a notification  of
** the changes is  added. You  are allowed to  redistribute it  and sell  it,
** alone or as a part of another product.
**
** (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2000
**
** Changed in August 2001
*/

/* GENERAL thread-safe features */

/* Assumes:
  T2 -> INTEGER
  T3 -> CHARACTER
  T4 -> REAL
  T5 -> DOUBLE
  T6 -> BOOLEAN
  T8 -> POINTER (void*)
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

void* se_malloc(size_t size);


/*********\
* general *
\*********/

/*
void print_integer(int i)
*/
void print_integer(T2 a)
{
  int i = (int)a;

  printf("%d",i);
}

/*
void print_real(float r)
*/
void print_real(T4 a)
{
  float r = (float)a;

  printf("%g",(double)r);
}

/*
void print_double(double d)
*/
void print_double(T4 a)
{
  double d = (double)a;

  printf("%g",d);
}

/*
void print_character(char c)
*/
void print_character(T3 a)
{
  char c = (char)a;

  printf("%c",c);
}

/*
void print_string(char *str)
*/
void print_string(T8 a)
{
  char *str = (char *)a;

  printf("%s",str);
}

/*
void print_line(char *str)
*/
void print_line(T8 a)
{
  char *str = (char *)a;

  printf("%s\n",str);
}

void print_newline(void)
{
  printf("\n");
}

void print_tab(void)
{
  printf("\t");
}

void print_spaces(int n)
{
  int i;

  for(i = 0;i < n;i++)
    printf(" ");
}

#define MAX_LINE_SIZE 1024

static pthread_key_t _get_line_key_;
static pthread_once_t _get_line_key_once_ = PTHREAD_ONCE_INIT;

static void _init_get_line_key_(void)
{
  pthread_key_create(&_get_line_key_,NULL);
}

int c_max_line_size(void)
{
  return MAX_LINE_SIZE;
}

T8 c_get_line(int max_size)
{
  char *str;
  int i,c;

  pthread_once(&_get_line_key_once_,_init_get_line_key_);

  str = (char *)pthread_getspecific(_get_line_key_);
  if (str == NULL)
  {
    str = se_malloc(MAX_LINE_SIZE);
    pthread_setspecific(_get_line_key_,(void *)str);
  }

  flockfile(stdin);
  i = 0;
  c = getchar_unlocked();
  while(c != '\n' && c != EOF && i < max_size-1)
  {
    str[i++] = c;
    c = getchar_unlocked();
  }
  funlockfile(stdin);
  str[i] = '\0';

  return (T8)str;
}

static void _get_token_(char *str,int max_size)
{
  int i,c;

  i = 0;
  flockfile(stdin);
  c = getchar_unlocked();
  while(c == '\n' || c == ' ' || c == '\t')
    c = getchar_unlocked();
  while(c != '\n' && c != ' ' && c != '\t' &&
        c != EOF && i < max_size-1)
  {
    str[i++] = c;
    c = getchar_unlocked();
  }
  funlockfile(stdin);
  str[i] = '\0';
}

int c_get_integer(void)
{
  int i;
  char str[16];

  _get_token_(str,16);
  i = atoi(str);

  return i;
}

float c_get_float(void)
{
  double f;
  char str[32];

  _get_token_(str,32);
  f = atof(str);

  return (float)f;
}

double c_get_double(void)
{
  double d;

  char str[32];

  _get_token_(str,32);
  d = atof(str);

  return d;
}

char c_get_character(void)
{
  int c;

  flockfile(stdin);
  c = getchar_unlocked();
  funlockfile(stdin);

  return (char)c;
}

void c_wait_newline(void)
{
  flockfile(stdin);
  while(getchar_unlocked() != '\n')
    ;
  funlockfile(stdin);
}

void c_second_sleep(int sec)
{
  sleep(sec);
}

void c_microsecond_sleep(int usec)
{
  usleep((unsigned long)usec);
}

void c_nanosecond_sleep(int nsec)
{
  struct timespec tm;

  tm.tv_sec = 0;
  tm.tv_nsec = (long)nsec;

  nanosleep(&tm,NULL);
}

int get_current_time(void)
{
  return time(NULL);
}

static int *rnd_seed = NULL;
static pthread_once_t _rnd_seed_once_init_ = PTHREAD_ONCE_INIT;

static void _init_random_generator_(void)
{
  unsigned long long t;

  rnd_seed = se_malloc(sizeof(int));
  asm volatile("rdtsc": "=A" (t));
  *rnd_seed = (int)((unsigned int)t & 0xFFFFU);
}

int maximum_random_integer(void)
{
  return RAND_MAX;
}

int random_integer(int max)
{
  int i,r;

  pthread_once(&_rnd_seed_once_init_,_init_random_generator_);

  i = rand_r(rnd_seed);
  if ((i < 0))
    i = -i;
  r=(int) ((double)((double)max+1.0)*(double)i/((double)RAND_MAX+1.0));

  return r;
}


/*************************\
* thread / thread_control *
\*************************/

/* SE-thread-safe uses [and abuses] read-write-locks.                 */
/* Those is the needed internal implementation (with write priority). */

#ifdef __USE_UNIX98
rwlock_t *_rwl_init_(void)
{
  rwlock_t *rwl;
  pthread_rwlockattr_t attr;
  int err;

  rwl = (rwlock_t *)se_malloc(sizeof(pthread_rwlock_t));
  err = pthread_rwlockattr_init(&attr);
  err = err |
	pthread_rwlockattr_setkind_np(&attr,PTHREAD_RWLOCK_PREFER_WRITER_NP);
  err = err |
        pthread_rwlock_init(rwl,&attr);

  if (err)
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_init_\n");
    exit(1);
  }

  return rwl;
}

void _rwl_read_lock_(rwlock_t *rwl)
{
  if (pthread_rwlock_rdlock(rwl))
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_read_lock_\n");
    exit(1);
  }
}

void _rwl_read_unlock_(rwlock_t *rwl)
{
  if (pthread_rwlock_unlock(rwl))
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_read_unlock_\n");
    exit(1);
  }
}

void _rwl_write_lock_(rwlock_t *rwl)
{
  if (pthread_rwlock_wrlock(rwl))
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_read_lock_\n");
    exit(1);
  }
}

void _rwl_write_unlock_(rwlock_t *rwl)
{
  if (pthread_rwlock_unlock(rwl))
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_write_unlock_\n");
    exit(1);
  }
}
#else
rwlock_t *_rwl_init_(void)
{
  rwlock_t *rwl;
  int err;

  rwl = (rwlock_t *)se_malloc(sizeof(rwlock_t));
  rwl->r_active = 0;
  rwl->w_active = 0;
  rwl->r_wait = 0;
  rwl->w_wait = 0;
  err = pthread_mutex_init (&rwl->mutex, NULL);
  err = err | pthread_cond_init (&rwl->read, NULL);
  err = err | pthread_cond_init (&rwl->write, NULL);

  if (err)
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_init_\n");
    exit(1);
  }

  return rwl;
}

void _rwl_read_lock_(rwlock_t *rwl)
{
  int err;

  err = pthread_mutex_lock(&rwl->mutex);
  if (rwl->w_active || rwl->w_wait > 0)
  {
    rwl->r_wait++;
    while (rwl->w_active || rwl->w_wait > 0)
    {
      err = err | pthread_cond_wait (&rwl->read, &rwl->mutex);
      if (err != 0)
        break;
    }
    rwl->r_wait--;
  }
  rwl->r_active++;
  err = err | pthread_mutex_unlock (&rwl->mutex);

  if (err)
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_read_lock_\n");
    exit(1);
  }
}

void _rwl_read_unlock_(rwlock_t *rwl)
{
  int err;

  err = pthread_mutex_lock(&rwl->mutex);
  rwl->r_active--;
  if (rwl->r_active == 0 && rwl->w_wait > 0)
    err = err | pthread_cond_signal (&rwl->write);
  err = err | pthread_mutex_unlock (&rwl->mutex);

  if (err)
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_read_lock_\n");
    exit(1);
  }
}

void _rwl_write_lock_(rwlock_t *rwl)
{
  int err;

  err = pthread_mutex_lock(&rwl->mutex);
  if (rwl->w_active || rwl->r_active > 0)
  {
    rwl->w_wait++;
    while (rwl->w_active || rwl->r_active > 0)
    {
      err = err | pthread_cond_wait (&rwl->write,&rwl->mutex);
      if (err != 0)
        break;
    }
    rwl->w_wait--;
  }
  rwl->w_active = 1;
  err = err | pthread_mutex_unlock (&rwl->mutex);

  if (err)
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_write_lock_\n");
    exit(1);
  }
}

void _rwl_write_unlock_(rwlock_t *rwl)
{
  int err;

  err = pthread_mutex_lock(&rwl->mutex);
  rwl->w_active = 0;
  if (rwl->w_wait > 0)
    err = err | pthread_cond_signal(&rwl->write);
  else if (rwl->r_wait > 0)
    err = err | pthread_cond_broadcast(&rwl->read);
  err = err | pthread_mutex_unlock (&rwl->mutex);

  if (err)
  {
    fprintf(stderr,"INTERNAL ERROR: _rwl_write_unlock_\n");
    exit(1);
  }
}
#endif

rwlock_t *_thread_creation_tree_rwl_ = NULL;
typedef void (*CMainFunction)(void*,void*);

pthread_key_t _thread_data_key_;

typedef struct
{
  CMainFunction rout;
  void *obj;
  _Thread_Data_ *td;
} _ThreadStubData_;


void _add_new_thread_(_Thread_Data_ *td)
  /* called at the start of the new thread */
  /* this routine sets thread id entry and manages thread creation tree */
{
  int err = 0;

  if (td == NULL) /* main thread */
  {
    if (pthread_key_create(&_thread_data_key_,NULL))
      err = 1;

    _thread_creation_tree_rwl_ = _rwl_init_();

    td = (_Thread_Data_ *)se_malloc(sizeof(_Thread_Data_));

    td->id = pthread_self();

    td->running = 1;
    td->finish = 0;
    td->detached = 0;
    td->error = err;

    td->parent = NULL;
    td->first_child = NULL;
    td->sibling = NULL;

#ifndef _BOOST_MODE_
    td->_se_dst = NULL;
#endif    

    td->rwl = _rwl_init_();

    if (pthread_setspecific(_thread_data_key_,(void*)td))
      td->error = 1;
  }
  else
  {
    td->id = pthread_self();

    if (pthread_setspecific(_thread_data_key_,(void*)td))
      td->error = 1;

    _rwl_write_lock_(_thread_creation_tree_rwl_);
    if (td->parent->first_child == NULL)
      td->parent->first_child = td;
    else
    {
      _Thread_Data_ *t;

      for(t=td->parent->first_child;t->sibling != NULL;t=t->sibling)
        ;
      t->sibling = td;
    }

#ifndef _BOOST_MODE_
    td->_se_dst = NULL;
#endif    

    _rwl_write_unlock_(_thread_creation_tree_rwl_);
  }
}

void *start_thread_stub(void*a)
{
  _ThreadStubData_ *stub_data = (_ThreadStubData_ *)a;

  _add_new_thread_(stub_data->td);
  stub_data->rout(stub_data->obj,stub_data->td);

  return 0;
}

_Thread_Data_ *initialize_new_pthread(void)
  /* to call before the start of a new thread, by its parent */
{
  _Thread_Data_ *td,*parent_td;

  td = (_Thread_Data_ *)se_malloc(sizeof(_Thread_Data_));

  td->running = 0;
  td->finish = 0;
  td->detached = 0;
  td->error = 0;

  parent_td = (_Thread_Data_*)pthread_getspecific(_thread_data_key_);

  if (parent_td == NULL)
  {
    fprintf(stderr,"INTERNAL ERROR: initialize_new_pthread\n");
    exit(1);
  }

  td->parent = parent_td;
  td->first_child = NULL;
  td->sibling = NULL;

  td->rwl = _rwl_init_();

  return td;
}

/*
_Thread_Data_ *get_current_thread_data(void)
*/
T8 get_current_thread_data(void)
{
  return (_Thread_Data_*)pthread_getspecific(_thread_data_key_);
}

/*
int is_root_thread(_Thread_Data_ *td)
*/
T6 is_root_thread(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  T6 result;

  _rwl_read_lock_(td->rwl);
  result = (td->parent == NULL ? 1 : 0);
  _rwl_read_unlock_(td->rwl);

  return result;
}

/*
void finalize_pthread_state(_Thread_Data_ *td)
*/
void finalize_pthread_state(T8 a)
{
/*
  _Thread_Data_ *td = (_Thread_Data_*)a;

  rwl_destroy(td->rwl);
  td->rwl = NULL;
*/
}

/*
int pthread_running(_Thread_Data_ *td)
*/
T6 pthread_running(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  T6 result = 0;

  if (td != NULL)
  {
    _rwl_read_lock_(td->rwl);
    result = td->running;
    _rwl_read_unlock_(td->rwl);
  }

  return result;
}

void set_pthread_running_state(T8 a,T6 st)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;

  _rwl_write_lock_(td->rwl);
  td->running = st;
  _rwl_write_unlock_(td->rwl);
}

/*
int pthread_finish(_Thread_Data_ *td)
*/
T6 pthread_finish(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  T6 result = 0;

  if (td != NULL)
  {
    _rwl_read_lock_(td->rwl);
    result = td->finish;
    _rwl_read_unlock_(td->rwl);
  }

  return result;
}

void set_pthread_finish_state(T8 a,T6 st)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;

  _rwl_write_lock_(td->rwl);
  td->finish = st;
  _rwl_write_unlock_(td->rwl);
}

/*
int pthread_detached(_Thread_Data_ *td)
*/
T6 pthread_detached(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  T6 result;

  _rwl_read_lock_(td->rwl);
  result = td->detached;
  _rwl_read_unlock_(td->rwl);

  return result;
}

/*
int pthread_error(_Thread_Data_ *td)
*/
T6 pthread_error(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  T6 result;

  if (td == NULL)
    return 0;

  _rwl_read_lock_(td->rwl);
  result = td->error;
  _rwl_read_unlock_(td->rwl);

  return result;
}


/*
void start_pthread(void *obj,CMainFunction rout)
*/
void start_pthread(void *obj,T8 a)
{
  CMainFunction rout = (CMainFunction)a;
  _ThreadStubData_ *stub_data;

  stub_data = (_ThreadStubData_ *)se_malloc(sizeof(_ThreadStubData_));
  stub_data->rout = rout;
  stub_data->obj = obj;
  stub_data->td = initialize_new_pthread();
  _rwl_write_lock_(stub_data->td->rwl);
  if (pthread_create(&(stub_data->td->id),NULL,&start_thread_stub,stub_data))
    stub_data->td->error = 1;
  _rwl_write_unlock_(stub_data->td->rwl);
}

/*
void start_detached_pthread(void *obj,CMainFunction rout)
*/
void start_detached_pthread(void *obj,T8 a)
{
  CMainFunction rout = (CMainFunction)a;
  _ThreadStubData_ *stub_data;

  stub_data = (_ThreadStubData_ *)se_malloc(sizeof(_ThreadStubData_));
  stub_data->rout = rout;
  stub_data->obj = obj;
  stub_data->td = initialize_new_pthread();
  _rwl_write_lock_(stub_data->td->rwl);
  if (pthread_create(&(stub_data->td->id),NULL,&start_thread_stub,stub_data))
    stub_data->td->error = 1;
  else if (pthread_detach(stub_data->td->id))
      stub_data->td->error = 1;
  stub_data->td->detached = 1;
  _rwl_write_unlock_(stub_data->td->rwl);
}

/*
int is_same_thread(_Thread_Data_ *td)
*/
T6 is_same_thread(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  T6 result;

  _rwl_read_lock_(td->rwl);
  result = (pthread_equal(td->id,pthread_self()) ? 1 : 0);
  _rwl_read_unlock_(td->rwl);

  return result;
}

/*
void detach_pthread(_Thread_Data_ *td)
*/
void detach_pthread(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;

  _rwl_write_lock_(td->rwl);
  if (pthread_detach(td->id))
    td->error = 1;
  td->detached = 1;
  _rwl_write_unlock_(td->rwl);
}

/*
void exit_pthread(_Thread_Data_ *td)
*/
void exit_pthread(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  void *res;

  _rwl_write_lock_(td->rwl);
  pthread_exit(res);
  _rwl_write_unlock_(td->rwl);
  finalize_pthread_state(a);
}

/*
void join_pthread(_Thread_Data_ *td)
*/
void join_pthread(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  pthread_t id;

  _rwl_read_lock_(td->rwl);
  id = td->id;
  _rwl_read_unlock_(td->rwl);

  if (pthread_join(id,NULL))
  {
    _rwl_write_lock_(td->rwl);
    td->error = 1;
    _rwl_write_unlock_(td->rwl);
  }
}

/*
void join_all_childs_pthread(_Thread_Data_ *td)
*/
void join_all_childs_pthread(T8 a)
{
  _Thread_Data_ *td = (_Thread_Data_*)a;
  _Thread_Data_ *t;

  _rwl_read_lock_(_thread_creation_tree_rwl_);
  for(t = td->first_child;t != NULL;t = t->sibling)
  {
    int finish,detached;

    _rwl_read_unlock_(_thread_creation_tree_rwl_);
    _rwl_read_lock_(t->rwl);
    finish = t->finish;
    detached = t->detached;
    _rwl_read_unlock_(t->rwl);

    if (!finish && !detached)
    {
      if (pthread_join(t->id,NULL))
      {
        _rwl_write_lock_(td->rwl);
        td->error = 1;
        _rwl_write_unlock_(td->rwl);
      }
    }
    _rwl_read_lock_(_thread_creation_tree_rwl_);
  }
  _rwl_read_unlock_(_thread_creation_tree_rwl_);
}

/*
int same_thread_as(_Thread_Data_ *td1,_Thread_Data_ *td2)
*/
T6 same_thread_as(T8 a,T8 b)
{
  _Thread_Data_ *td1 = (_Thread_Data_*)a;
  _Thread_Data_ *td2 = (_Thread_Data_*)b;

  return (pthread_equal(td1->id,td2->id) ? 1 : 0);
}


/*******\
* mutex *
\*******/

typedef struct
{
  pthread_mutex_t mutex;
  T6 error; /* 1: TRUE, 0: FALSE */
} MutexType;

/*
MutexType *initialize_mutex_state(void)
*/
T8 initialize_mutex_state(void)
{
  MutexType *mt;

  mt = (MutexType *)se_malloc(sizeof(MutexType));
  mt->error = 0;
  if (pthread_mutex_init(&mt->mutex,NULL))
    mt->error = 1;

  return mt;
}

/*
void finalize_mutex_state(MutexType *mt)
*/
void finalize_mutex_state(T8 a)
{
  MutexType *mt = (MutexType *)a;

  if (pthread_mutex_destroy(&mt->mutex))
    mt->error = 1;
}

/*
int mutex_error(MutexType *mt)
*/
T6 mutex_error(T8 a)
{
  MutexType *mt = (MutexType *)a;

  return (mt == NULL ? 0 : mt->error);
}

/*
void mutex_lock(MutexType *mt)
*/
void mutex_lock(T8 a)
{
  MutexType *mt = (MutexType *)a;

  if (pthread_mutex_lock(&mt->mutex))
    mt->error = 1;
}

/*
int mutex_trylock(MutexType *mt)
*/
T6 mutex_trylock(T8 a)
 /* returns: 0 on FAILURE and nonzero if lock succeeds */
{
  MutexType *mt = (MutexType *)a;

  int status;

  status = pthread_mutex_trylock(&mt->mutex);
  if (status == EBUSY)
    return 0;
  if (status != 0)
  {
    mt->error = 1;
    return 0;
  }
  return 1;
}

/*
void mutex_unlock(MutexType *mt)
*/
void mutex_unlock(T8 a)
{
  MutexType *mt = (MutexType *)a;

  if (pthread_mutex_unlock(&mt->mutex))
    mt->error = 1;
}


/********************\
* condition_variable *
\********************/

typedef struct
{
  pthread_cond_t cond;
  T6 error; /* 1: TRUE, 0: FALSE */
} CondVarType;

/*
CondVarType *initialize_cond_var_state(void)
*/
T8 initialize_cond_var_state(void)
{
  CondVarType *cvt;

  cvt = (CondVarType *)se_malloc(sizeof(CondVarType));
  cvt->error = 0;
  if (pthread_cond_init(&cvt->cond,NULL))
    cvt->error = 1;

  return cvt;
}

/*
void finalize_cond_var_state(CondVarType *cvt)
*/
void finalize_cond_var_state(T8 a)
{
  CondVarType *cvt = (CondVarType *)a;

  if (pthread_cond_destroy(&cvt->cond))
    cvt->error = 1;
}

/*
int cond_var_error(CondVarType *cvt)
*/
T6 cond_var_error(T8 a)
{
  CondVarType *cvt = (CondVarType *)a;

  return (cvt == NULL ? 0 : cvt->error);
}

/*
void cond_var_wait(CondVarType *cvt,MutexType *mt)
*/
void cond_var_wait(T8 a,T8 b)
{
  CondVarType *cvt = (CondVarType *)a;
  MutexType *mt = (MutexType *)b;

  if (pthread_cond_wait(&cvt->cond,&mt->mutex))
    cvt->error = 1;
}

/*
int cond_var_timedwait(CondVarType *cvt,MutexType *mt,int timeout)
*/
T6 cond_var_timedwait(T8 a,T8 b,int timeout)
  /* timeout is absolute time in seconds */
  /* returns: 0 on timeout and nonzero if signaled */
{
  CondVarType *cvt = (CondVarType *)a;
  MutexType *mt = (MutexType *)b;
  struct timespec t;
  int status;
  T6 result;

/*
  t.tv_sec = time(NULL) + timeout;
*/
  t.tv_sec = timeout;
  t.tv_nsec = 0;
  status = pthread_cond_timedwait(&cvt->cond,&mt->mutex,&t);
  if (status == ETIMEDOUT)
    result = 0;
  else if (status)
  {
    result = 0;
    cvt->error = 1;
  }
  else
    result = 1;

  return result;
}

/*
void cond_var_signal(CondVarType *cvt)
*/
void cond_var_signal(T8 a)
{
  CondVarType *cvt = (CondVarType *)a;

  if (pthread_cond_signal(&cvt->cond))
    cvt->error = 1;
}

/*
void cond_var_broadcast(CondVarType *cvt)
*/
void cond_var_broadcast(T8 a)
{
  CondVarType *cvt = (CondVarType *)a;

  if (pthread_cond_broadcast(&cvt->cond))
    cvt->error = 1;
}


#ifdef __USE_UNIX98
/*****************\
* read_write_lock *
\*****************/

typedef struct
{
  pthread_rwlock_t rwl;
  T6 w_pref; /* 1: writer priority, 0: readers priority locking */
  T6 error; /* 1: TRUE, 0: FALSE */
} ReadWriteLockType;

/*
ReadWriteLockType *initialize_rwl_state_with_write_priority(void)
*/
T8 initialize_rwl_state_with_write_priority(void)
{
  ReadWriteLockType *rwlt;
  pthread_rwlockattr_t attr;

  rwlt = (ReadWriteLockType *)se_malloc(sizeof(ReadWriteLockType));
  rwlt->error = 0;
  rwlt->w_pref = 1;
  if (pthread_rwlockattr_init(&attr))
    rwlt->error = 1;
  if (pthread_rwlockattr_setkind_np(&attr,PTHREAD_RWLOCK_PREFER_WRITER_NP))
    rwlt->error = 1;
  if (pthread_rwlock_init(&rwlt->rwl,&attr))
    rwlt->error = 1;

  return rwlt;
}

/*
ReadWriteLockType *initialize_rwl_state_with_read_priority(void)
*/
T8 initialize_rwl_state_with_read_priority(void)
{
  ReadWriteLockType *rwlt;
  pthread_rwlockattr_t attr;

  rwlt = (ReadWriteLockType *)se_malloc(sizeof(ReadWriteLockType));
  rwlt->error = 0;
  rwlt->w_pref = 0;
  if (pthread_rwlockattr_init(&attr))
    rwlt->error = 1;
  if (pthread_rwlockattr_setkind_np(&attr,PTHREAD_RWLOCK_PREFER_READER_NP))
    rwlt->error = 1;
  if (pthread_rwlock_init(&rwlt->rwl,&attr))
    rwlt->error = 1;

  return rwlt;
}

/*
void finalize_rwl_state(ReadWriteLockType *rwlt)
*/
void finalize_rwl_state(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  if (pthread_rwlock_destroy(&rwlt->rwl))
    rwlt->error = 1;
}

/*
int rwl_error(ReadWriteLockType *rwlt)
*/
T6 rwl_error(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  return (rwlt == NULL ? 0 : rwlt->error);
}

/*
void c_rwl_readlock(ReadWriteLockType *rwlt)
*/
void c_rwl_readlock(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  if (pthread_rwlock_rdlock(&rwlt->rwl))
    rwlt->error = 1;
}

/*
int c_rwl_readtrylock(ReadWriteLockType *rwlt)
*/
T6 c_rwl_readtrylock(T8 a)
 /* returns: 0 on FAILURE and nonzero if lock succeeds */
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  int status;

  status = pthread_rwlock_tryrdlock(&rwlt->rwl);
  if (status == EBUSY)
    return 0;
  if (status != 0)
  {
    rwlt->error = 1;
    return 0;
  }
  return 1;
}

/*
void c_rwl_readunlock(ReadWriteLockType *rwlt)
*/
void c_rwl_readunlock(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  if (pthread_rwlock_unlock(&rwlt->rwl))
    rwlt->error = 1;
}

/*
void c_rwl_writelock(ReadWriteLockType *rwlt)
*/
void c_rwl_writelock(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  if (pthread_rwlock_wrlock(&rwlt->rwl))
    rwlt->error = 1;
}

/*
int c_rwl_writetrylock(ReadWriteLockType *rwlt)
*/
T6 c_rwl_writetrylock(T8 a)
 /* returns: 0 on FAILURE and nonzero if lock succeeds */
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  int status;

  status = pthread_rwlock_trywrlock(&rwlt->rwl);
  if (status == EBUSY)
    return 0;
  if (status != 0)
  {
    rwlt->error = 1;
    return 0;
  }
  return 1;
}

/*
void c_rwl_writeunlock(ReadWriteLockType *rwlt)
*/
void c_rwl_writeunlock(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  if (pthread_rwlock_unlock(&rwlt->rwl))
    rwlt->error = 1;
}

/*
int c_write_lock_priority(ReadWriteLockType *rwlt)
*/
T6 c_write_lock_priority(T8 a)
{
  ReadWriteLockType *rwlt = (ReadWriteLockType *)a;

  return rwlt->w_pref;
}
#endif

